# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

from scenario import Context, State

import ops

META = {
    'name': 'traced_charm',
    'requires': {'charm-tracing': {'interface': 'tracing', 'limit': 1}},
}


class TracedCharm(ops.CharmBase):
    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        self.tracing = ops.tracing.Tracing(self, 'charm-tracing')
        for evt in self.on.events().values():
            self.framework.observe(evt, self._on_event)

    def _on_event(self, event: ops.EventBase):
        pass


def test_trace_data():
    ctx = Context(TracedCharm, meta=META)
    ctx.run(ctx.on.start(), State(leader=True))

    assert {s.name for s in ctx.trace_data} == {
        # The entry point and root span.
        'ops.main',
        # The first-party charm library.
        'ops.tracing.Tracing',
        # Start event emitted on this charm.
        'start: TracedCharm',
        # Start event emitted on the first party charm lib.
        'start: Tracing',
        # Emitted on the leader.
        'collect_app_status: TracedCharm',
        # Emitted on all units.
        'collect_unit_status: TracedCharm',
    }

    main_span = next(s for s in ctx.trace_data if s.name == 'ops.main')
    assert not main_span.parent
    assert {e.name for e in main_span.events} == {
        'StartEvent',
        'CollectStatusEvent',
        'PreCommitEvent',
        'CommitEvent',
    }



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import pytest
from scenario import Context
from scenario.state import (
    Address,
    BindAddress,
    Network,
    Relation,
    State,
    SubordinateRelation,
)

from ops import RelationNotFoundError
from ops.charm import CharmBase
from ops.framework import Framework


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(CharmBase):
        _call = None
        called = False

        def __init__(self, framework: Framework):
            super().__init__(framework)

            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if MyCharm._call:
                MyCharm.called = True
                MyCharm._call(self, event)

    return MyCharm


def test_ip_get(mycharm):
    ctx = Context(
        mycharm,
        meta={
            'name': 'foo',
            'requires': {
                'metrics-endpoint': {'interface': 'foo'},
                'deadnodead': {'interface': 'bar'},
            },
            'extra-bindings': {'foo': {}},
        },
    )

    with ctx(
        ctx.on.update_status(),
        State(
            relations=[
                Relation(
                    interface='foo',
                    remote_app_name='remote',
                    endpoint='metrics-endpoint',
                    id=1,
                ),
            ],
            networks={Network('foo', [BindAddress([Address('4.4.4.4')])])},
        ),
    ) as mgr:
        # we have a network for the relation
        rel = mgr.charm.model.get_relation('metrics-endpoint')
        assert str(mgr.charm.model.get_binding(rel).network.bind_address) == '192.0.2.0'

        # we have a network for a binding without relations on it
        assert str(mgr.charm.model.get_binding('deadnodead').network.bind_address) == '192.0.2.0'

        # and an extra binding
        assert str(mgr.charm.model.get_binding('foo').network.bind_address) == '4.4.4.4'


def test_no_sub_binding(mycharm):
    ctx = Context(
        mycharm,
        meta={
            'name': 'foo',
            'requires': {'bar': {'interface': 'foo', 'scope': 'container'}},
        },
    )

    with ctx(
        ctx.on.update_status(),
        State(
            relations=[
                SubordinateRelation('bar'),
            ]
        ),
    ) as mgr:
        with pytest.raises(RelationNotFoundError):
            # sub relations have no network
            mgr.charm.model.get_binding('bar').network  # noqa: B018  # Used to trigger the error.


def test_no_relation_error(mycharm):
    """Attempting to call get_binding on a non-existing relation -> RelationNotFoundError"""

    ctx = Context(
        mycharm,
        meta={
            'name': 'foo',
            'requires': {'metrics-endpoint': {'interface': 'foo'}},
            'extra-bindings': {'bar': {}},
        },
    )

    with ctx(
        ctx.on.update_status(),
        State(
            relations=[
                Relation(
                    interface='foo',
                    remote_app_name='remote',
                    endpoint='metrics-endpoint',
                    id=1,
                ),
            ],
            networks={Network('bar')},
        ),
    ) as mgr:
        with pytest.raises(RelationNotFoundError):
            mgr.charm.model.get_binding('foo').network  # noqa: B018  # Used to trigger the error.


def test_juju_info_network_default(mycharm):
    ctx = Context(
        mycharm,
        meta={'name': 'foo'},
    )

    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        # we have a network for the relation
        assert str(mgr.charm.model.get_binding('juju-info').network.bind_address) == '192.0.2.0'


def test_explicit_juju_info_network_override(mycharm):
    ctx = Context(
        mycharm,
        meta={
            'name': 'foo',
            # this charm for whatever reason explicitly defines a juju-info endpoint
            'requires': {'juju-info': {'interface': 'juju-info'}},
        },
    )

    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        assert mgr.charm.model.get_binding('juju-info').network.bind_address



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import os
from tempfile import TemporaryDirectory

import pytest
from scenario import ActiveStatus, Context
from scenario._runtime import Runtime, UncaughtCharmError
from scenario.state import Relation, State, _CharmSpec, _Event

import ops
from ops._main import _Abort


def charm_type():
    class _CharmEvents(ops.CharmEvents):
        pass

    class MyCharm(ops.CharmBase):
        on = _CharmEvents()  # type: ignore
        _event = None

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._catchall)

        def _catchall(self, e: ops.EventBase):
            if self._event:
                return
            MyCharm._event = e

    return MyCharm


def test_event_emission():
    with TemporaryDirectory():
        meta = {
            'name': 'foo',
            'requires': {'ingress-per-unit': {'interface': 'ingress_per_unit'}},
        }

        my_charm_type = charm_type()

        class MyEvt(ops.EventBase):
            pass

        my_charm_type.on.define_event('bar', MyEvt)

        runtime = Runtime(
            'foo',
            _CharmSpec(
                my_charm_type,
                meta=meta,
            ),
        )

        with runtime.exec(
            state=State(),
            event=_Event('bar'),
            context=Context(my_charm_type, meta=meta),
        ) as manager:
            manager.run()

        assert my_charm_type._event
        assert isinstance(my_charm_type._event, MyEvt)


@pytest.mark.parametrize('app_name', ('foo', 'bar-baz', 'QuX2'))
@pytest.mark.parametrize('unit_id', (1, 2, 42))
def test_unit_name(app_name, unit_id):
    meta = {
        'name': app_name,
    }

    my_charm_type = charm_type()

    runtime = Runtime(
        app_name,
        _CharmSpec(
            my_charm_type,
            meta=meta,
        ),
        unit_id=unit_id,
    )

    with runtime.exec(
        state=State(),
        event=_Event('start'),
        context=Context(my_charm_type, meta=meta),
    ) as manager:
        assert manager.charm.unit.name == f'{app_name}/{unit_id}'


def test_env_clean_on_charm_error(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv('SCENARIO_BARE_CHARM_ERRORS', 'false')
    meta = {'name': 'frank', 'requires': {'box': {'interface': 'triangle'}}}

    my_charm_type = charm_type()

    runtime = Runtime(
        'frank',
        _CharmSpec(
            my_charm_type,
            meta=meta,
        ),
    )

    remote_name = 'ava'
    rel = Relation('box', remote_app_name=remote_name)
    with pytest.raises(UncaughtCharmError) as exc:
        with runtime.exec(
            state=State(relations={rel}),
            event=_Event('box_relation_changed', relation=rel),
            context=Context(my_charm_type, meta=meta),
        ) as manager:
            assert manager._juju_context.remote_app_name == remote_name
            assert 'JUJU_REMOTE_APP' in os.environ
            _ = 1 / 0  # raise some error
    # Ensure that some other error didn't occur (like AssertionError!).
    assert 'ZeroDivisionError' in str(exc.value)

    # Ensure that the Juju environment didn't leak into the outside one.
    assert os.getenv('JUJU_REMOTE_APP', None) is None


def test_juju_version_is_set_in_environ():
    version = '2.9'

    class MyCharm(ops.CharmBase):
        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            framework.observe(self.on.start, self._on_start)

        def _on_start(self, _: ops.StartEvent):
            with pytest.warns(DeprecationWarning):
                assert ops.JujuVersion.from_environ() == version

    ctx = Context(MyCharm, meta={'name': 'foo'}, juju_version=version)
    ctx.run(ctx.on.start(), State())


@pytest.mark.parametrize('exit_code', (-1, 0, 1, 42))
def test_ops_raises_abort(exit_code: int, monkeypatch: pytest.MonkeyPatch):
    class MyCharm(ops.CharmBase):
        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            framework.observe(self.on.start, self._on_start)

        def _on_start(self, _: ops.StartEvent):
            self.unit.status = ops.ActiveStatus()
            # Charms can't actually do this (_Abort is private), but this is
            # simpler than causing the framework to raise it.
            raise _Abort(exit_code)

    monkeypatch.setenv('SCENARIO_BARE_CHARM_ERRORS', 'false')
    ctx = Context(MyCharm, meta={'name': 'foo'})
    if exit_code == 0:
        state_out = ctx.run(ctx.on.start(), State())
        assert {e.handle.kind for e in ctx.emitted_events} == {'start'}
        assert state_out.unit_status == ActiveStatus()
    else:
        with pytest.raises(UncaughtCharmError) as exc:
            ctx.run(ctx.on.start(), State())
        assert isinstance(exc.value.__cause__, _Abort)
        assert exc.value.__cause__.exit_code == exit_code


class ValueErrorCharm(ops.CharmBase):
    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        framework.observe(self.on.update_status, self._on_update_status)

    def _on_update_status(self, event: ops.EventBase):
        raise ValueError()


@pytest.mark.parametrize(
    ('expected_error', 'bare_charm_errors'),
    (
        # any non-zero string that `isdigit`
        (ValueError, '1'),
        (ValueError, '007'),
        # 'true' (case-insensitive)
        (ValueError, 'true'),
        (ValueError, 'True'),
        (ValueError, 'tRuE'),
        # falsey digits
        (UncaughtCharmError, '0'),
        (UncaughtCharmError, '00'),
        # any other value
        (UncaughtCharmError, '-1'),
        (UncaughtCharmError, 'false'),
        (UncaughtCharmError, 'yes'),
        (UncaughtCharmError, '✩ anything ✨ else ✧'),
        # the actually unset case is tested separately
        # the empty string is treated as 'any other value'
        (UncaughtCharmError, ''),
    ),
)
def test_bare_charm_errors_set(
    monkeypatch: pytest.Monkeypatch, expected_error: type[Exception], bare_charm_errors: str | None
):
    monkeypatch.setenv('SCENARIO_BARE_CHARM_ERRORS', bare_charm_errors)
    ctx = Context(ValueErrorCharm, meta={'name': 'value-error'})
    with pytest.raises(expected_error):
        ctx.run(ctx.on.update_status(), State())


def test_bare_charm_errors_not_set(monkeypatch: pytest.Monkeypatch):
    monkeypatch.delenv('SCENARIO_BARE_CHARM_ERRORS', raising=False)
    ctx = Context(ValueErrorCharm, meta={'name': 'value-error'})
    with pytest.raises(UncaughtCharmError):
        ctx.run(ctx.on.update_status(), State())



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import copy
import tempfile
from collections.abc import Callable, Generator, Iterable
from dataclasses import asdict, replace
from typing import Any

import pytest
import yaml
from scenario.context import Context
from scenario.state import (
    _DEFAULT_JUJU_DATABAG,
    Address,
    BindAddress,
    CheckInfo,
    CloudCredential,
    CloudSpec,
    Container,
    Model,
    Network,
    Notice,
    PeerRelation,
    Relation,
    RelationBase,
    Resource,
    Secret,
    State,
    Storage,
    StoredState,
    SubordinateRelation,
    TCPPort,
    _Event,
    _next_storage_index,
    layer_from_rockcraft,
)

import ops
from ops.charm import CharmBase, CharmEvents, CollectStatusEvent
from ops.framework import EventBase, Framework
from ops.model import ActiveStatus, UnknownStatus, WaitingStatus
from tests.helpers import jsonpatch_delta, sort_patch, trigger

CUSTOM_EVT_SUFFIXES = {
    'relation_created',
    'relation_joined',
    'relation_changed',
    'relation_departed',
    'relation_broken',
    'storage_attached',
    'storage_detaching',
    'action',
    'pebble_ready',
    'pebble_custom_notice',
}


@pytest.fixture(scope='function')
def mycharm():
    class MyCharmEvents(CharmEvents):
        @classmethod
        def define_event(cls, event_kind: str, event_type: type[EventBase]):
            if getattr(cls, event_kind, None):
                delattr(cls, event_kind)
            return super().define_event(event_kind, event_type)

    class MyCharm(CharmBase):
        _call: Callable[[MyCharm, _Event], None] | None = None
        called = False
        on = MyCharmEvents()

        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if self._call:
                MyCharm.called = True
                self._call(event)

    return MyCharm


@pytest.fixture
def state():
    return State(config={'foo': 'bar'}, leader=True)


def test_bare_event(state, mycharm):
    out = trigger(
        state,
        'start',
        mycharm,
        meta={'name': 'foo'},
        config={'options': {'foo': {'type': 'string'}}},
    )
    out_purged = replace(out, stored_states=state.stored_states)
    assert jsonpatch_delta(state, out_purged) == []


def test_leader_get(state, mycharm):
    def pre_event(charm):
        assert charm.unit.is_leader()

    trigger(
        state,
        'start',
        mycharm,
        meta={'name': 'foo'},
        config={'options': {'foo': {'type': 'string'}}},
        pre_event=pre_event,
    )


def test_status_setting(state, mycharm):
    def call(charm: CharmBase, e):
        if isinstance(e, CollectStatusEvent):
            return

        assert isinstance(charm.unit.status, UnknownStatus)
        charm.unit.status = ActiveStatus('foo test')
        charm.app.status = WaitingStatus('foo barz')

    mycharm._call = call
    out = trigger(
        state,
        'start',
        mycharm,
        meta={'name': 'foo'},
        config={'options': {'foo': {'type': 'string'}}},
    )
    assert out.unit_status == ActiveStatus('foo test')
    assert out.app_status == WaitingStatus('foo barz')
    assert out.workload_version == ''

    # ignore stored state in the delta
    out_purged = replace(out, stored_states=state.stored_states)
    assert jsonpatch_delta(out_purged, state) == sort_patch([
        {'op': 'replace', 'path': '/app_status/message', 'value': 'foo barz'},
        {'op': 'replace', 'path': '/app_status/name', 'value': 'waiting'},
        {'op': 'replace', 'path': '/unit_status/message', 'value': 'foo test'},
        {'op': 'replace', 'path': '/unit_status/name', 'value': 'active'},
    ])


@pytest.mark.parametrize('connect', (True, False))
def test_container(connect, mycharm):
    def pre_event(charm: CharmBase):
        container = charm.unit.get_container('foo')
        assert container is not None
        assert container.name == 'foo'
        assert container.can_connect() is connect

    trigger(
        State(containers={Container(name='foo', can_connect=connect)}),
        'start',
        mycharm,
        meta={
            'name': 'foo',
            'containers': {'foo': {'resource': 'bar'}},
        },
        pre_event=pre_event,
    )


def test_relation_get(mycharm):
    def pre_event(charm: CharmBase):
        rel = charm.model.get_relation('foo')
        assert rel is not None
        assert rel.data[charm.app]['a'] == 'because'

        assert rel.data[rel.app]['a'] == 'b'
        assert rel.data[charm.unit]['c'] == 'd'

        for unit in rel.units:
            if unit is charm.unit:
                continue
            if unit.name == 'remote/1':
                assert rel.data[unit]['e'] == 'f'
            else:
                assert not rel.data[unit]

    state = State(
        relations={
            Relation(
                endpoint='foo',
                interface='bar',
                local_app_data={'a': 'because'},
                remote_app_name='remote',
                remote_app_data={'a': 'b'},
                local_unit_data={'c': 'd'},
                remote_units_data={0: {}, 1: {'e': 'f'}, 2: {}},
            )
        }
    )
    trigger(
        state,
        'start',
        mycharm,
        meta={
            'name': 'local',
            'requires': {'foo': {'interface': 'bar'}},
        },
        pre_event=pre_event,
    )


def test_relation_set(mycharm):
    def event_handler(charm: CharmBase, _):
        rel = charm.model.get_relation('foo')
        assert rel is not None
        rel.data[charm.app]['a'] = 'b'
        rel.data[charm.unit]['c'] = 'd'

        # this will NOT raise an exception because we're not in an event context!
        # we're right before the event context is entered in fact.
        with pytest.raises(ops.RelationDataAccessError):
            rel.data[rel.app]['a'] = 'b'
        with pytest.raises(ops.RelationDataAccessError):
            rel.data[charm.model.get_unit('remote/1')]['c'] = 'd'

        assert charm.unit.is_leader()

    def pre_event(charm: CharmBase):
        assert charm.model.get_relation('foo')
        assert charm.model.app.planned_units() == 4

        # this would NOT raise an exception because we're not in an event context!
        # we're right before the event context is entered in fact.
        # with pytest.raises(Exception):
        #     rel.data[rel.app]["a"] = "b"
        # with pytest.raises(Exception):
        #     rel.data[charm.model.get_unit("remote/1")]["c"] = "d"

    mycharm._call = event_handler
    relation = Relation(
        endpoint='foo',
        interface='bar',
        remote_app_name='remote',
        remote_units_data={1: {}, 4: {}},
    )
    state = State(
        leader=True,
        planned_units=4,
        relations={relation},
    )

    assert not mycharm.called
    out = trigger(
        state,
        event='start',
        charm_type=mycharm,
        meta={
            'name': 'foo',
            'requires': {'foo': {'interface': 'bar'}},
        },
        pre_event=pre_event,
    )
    assert mycharm.called

    assert asdict(out.get_relation(relation.id)) == asdict(
        replace(
            relation,
            local_app_data={'a': 'b'},
            local_unit_data={'c': 'd', **_DEFAULT_JUJU_DATABAG},
        )
    )
    assert out.get_relation(relation.id).local_app_data == {'a': 'b'}
    assert out.get_relation(relation.id).local_unit_data == {
        'c': 'd',
        **_DEFAULT_JUJU_DATABAG,
    }


def test_checkinfo_changeid_none():
    info = CheckInfo('foo', change_id=None)
    assert info.change_id, 'None should result in a random change_id'
    info2 = CheckInfo('foo')  # None is also the default.
    assert info.change_id != info2.change_id


@pytest.mark.parametrize('id', ('', '28'))
def test_checkinfo_changeid(id: str | None):
    info = CheckInfo('foo', change_id=ops.pebble.ChangeID(id))
    assert info.change_id == ops.pebble.ChangeID(id)


@pytest.mark.parametrize(
    'klass,num_args',
    [
        (State, (1,)),
        (Resource, (1,)),
        (Address, (0, 2)),
        (BindAddress, (0, 2)),
        (Network, (0, 3)),
    ],
)
def test_positional_arguments(klass, num_args):
    for num in num_args:
        args = (None,) * num
        with pytest.raises(TypeError):
            klass(*args)


def test_model_positional_arguments():
    with pytest.raises(TypeError):
        Model('', '')


def test_container_positional_arguments():
    with pytest.raises(TypeError):
        Container('', True)


def test_container_default_values():
    name = 'foo'
    container = Container(name)
    assert container.name == name
    assert container.can_connect is False
    assert container.layers == {}
    assert container.service_statuses == {}
    assert container.mounts == {}
    assert container.execs == frozenset()
    assert container.layers == {}
    assert container._base_plan == {}


def test_state_default_values():
    state = State()
    assert state.config == {}
    assert state.relations == frozenset()
    assert state.networks == frozenset()
    assert state.containers == frozenset()
    assert state.storages == frozenset()
    assert state.opened_ports == frozenset()
    assert state.secrets == frozenset()
    assert state.resources == frozenset()
    assert state.deferred == []
    assert isinstance(state.model, Model)
    assert state.leader is False
    assert state.planned_units == 1
    assert state.app_status == UnknownStatus()
    assert state.unit_status == UnknownStatus()
    assert state.workload_version == ''


def test_deepcopy_state():
    containers = [Container('foo'), Container('bar')]
    state = State(containers=containers)
    state_copy = copy.deepcopy(state)
    for container in state.containers:
        copied_container = state_copy.get_container(container.name)
        assert container.name == copied_container.name


def test_replace_state():
    containers = [Container('foo'), Container('bar')]
    state = State(containers=containers, leader=True)
    state2 = replace(state, leader=False)
    assert state.leader != state2.leader
    assert state.containers == state2.containers


@pytest.mark.parametrize(
    'component,attribute,required_args',
    [
        (CloudCredential, 'attributes', {'auth_type': 'foo'}),
        (Secret, 'tracked_content', {}),
        (Secret, 'latest_content', {'tracked_content': {'password': 'password'}}),
        (Secret, 'remote_grants', {'tracked_content': {'password': 'password'}}),
        (Relation, 'local_app_data', {'endpoint': 'foo'}),
        (Relation, 'local_unit_data', {'endpoint': 'foo'}),
        (Relation, 'remote_app_data', {'endpoint': 'foo'}),
        (SubordinateRelation, 'local_app_data', {'endpoint': 'foo'}),
        (SubordinateRelation, 'local_unit_data', {'endpoint': 'foo'}),
        (SubordinateRelation, 'remote_app_data', {'endpoint': 'foo'}),
        (SubordinateRelation, 'remote_unit_data', {'endpoint': 'foo'}),
        (PeerRelation, 'local_app_data', {'endpoint': 'foo'}),
        (PeerRelation, 'local_unit_data', {'endpoint': 'foo'}),
        (Notice, 'last_data', {'key': 'foo'}),
        (Container, 'layers', {'name': 'foo'}),
        (Container, 'service_statuses', {'name': 'foo'}),
        (Container, 'mounts', {'name': 'foo'}),
        (Container, 'notices', {'name': 'foo'}),
        (StoredState, 'content', {}),
    ],
)
def test_immutable_content_dict(
    component: type[object], attribute: str, required_args: dict[str, Any]
):
    content = {'foo': 'bar'}
    obj1 = component(**required_args, **{attribute: content})
    obj2 = component(**required_args, **{attribute: content})
    assert getattr(obj1, attribute) == getattr(obj2, attribute) == content
    assert getattr(obj1, attribute) is not getattr(obj2, attribute)
    content['baz'] = 'qux'
    assert getattr(obj1, attribute) == getattr(obj2, attribute) == {'foo': 'bar'}
    # This shouldn't be done in a charm test, since the attribute should be immutable,
    # but it's convenient to verify that the content is not connected.
    object.__setattr__(obj1, attribute, {'baz': 'qux'})
    assert getattr(obj1, attribute) == {'baz': 'qux'}
    assert getattr(obj2, attribute) == {'foo': 'bar'}


@pytest.mark.parametrize(
    'component,attribute,required_args',
    [
        (CloudCredential, 'redacted', {'auth_type': 'foo'}),
        (CloudSpec, 'ca_certificates', {'type': 'foo'}),
        (BindAddress, 'addresses', {}),
        (Network, 'bind_addresses', {'binding_name': 'foo'}),
        (Network, 'ingress_addresses', {'binding_name': 'foo'}),
        (Network, 'egress_subnets', {'binding_name': 'foo'}),
    ],
)
def test_immutable_content_list(
    component: type[object], attribute: str, required_args: dict[str, Any]
):
    content = ['foo', 'bar']
    obj1 = component(**required_args, **{attribute: content})
    obj2 = component(**required_args, **{attribute: content})
    assert getattr(obj1, attribute) == getattr(obj2, attribute) == content
    assert getattr(obj1, attribute) is not getattr(obj2, attribute)
    content.append('baz')
    assert getattr(obj1, attribute) == getattr(obj2, attribute) == ['foo', 'bar']
    # This shouldn't be done in a charm test, since the attribute should be immutable,
    # but it's convenient to verify that the content is not connected.
    object.__setattr__(obj1, attribute, ['baz', 'qux'])
    assert getattr(obj1, attribute) == ['baz', 'qux']
    assert getattr(obj2, attribute) == ['foo', 'bar']


@pytest.mark.parametrize(
    'component,attribute,required_args',
    [
        (Relation, 'remote_units_data', {'endpoint': 'foo'}),
        (PeerRelation, 'peers_data', {'endpoint': 'foo'}),
    ],
)
def test_immutable_content_dict_of_dicts(
    component: type[object], attribute: str, required_args: dict[str, Any]
):
    content = {0: {'foo': 'bar'}, 1: {'baz': 'qux'}}
    obj1 = component(**required_args, **{attribute: content})
    obj2 = component(**required_args, **{attribute: content})
    assert getattr(obj1, attribute) == getattr(obj2, attribute) == content
    assert getattr(obj1, attribute) is not getattr(obj2, attribute)
    content[0]['baz'] = 'quux'
    assert (
        getattr(obj1, attribute)
        == getattr(obj2, attribute)
        == {0: {'foo': 'bar'}, 1: {'baz': 'qux'}}
    )
    # This shouldn't be done in a charm test, since the attribute should be immutable,
    # but it's convenient to verify that the content is not connected.
    object.__setattr__(obj1, attribute, {0: {'foo': 'qux'}})
    assert getattr(obj1, attribute) == {0: {'foo': 'qux'}}
    assert getattr(obj2, attribute) == {0: {'foo': 'bar'}, 1: {'baz': 'qux'}}


@pytest.mark.parametrize(
    'obj_in,attribute,get_method,key_attr',
    [
        ({'foo': 'bar'}, 'config', '', ''),
        (Relation('rel'), 'relations', 'get_relation', 'id'),
        (PeerRelation('peer'), 'relations', 'get_relation', 'id'),
        (SubordinateRelation('sub'), 'relations', 'get_relation', 'id'),
        (Network('foo'), 'networks', 'get_network', 'binding_name'),
        (Container('foo'), 'containers', 'get_container', 'name'),
        (Storage('foo'), 'storages', 'get_storage', 'name'),
        (TCPPort(80), 'opened_ports', '', ''),
        (Secret({'foo': 'bar'}), 'secrets', '', ''),
        (Resource(name='foo', path='bar'), 'resources', '', ''),
        (StoredState(), 'stored_states', 'get_stored_state', 'name'),
    ],
)
def test_state_immutable(obj_in, attribute: str, get_method: str, key_attr: str, mycharm):
    state_in = State(**{attribute: obj_in if isinstance(obj_in, dict) else [obj_in]})

    state_out: State = trigger(
        state_in,
        event='start',
        charm_type=mycharm,
        meta={
            'name': 'foo',
            'containers': {'foo': {'resource': 'bar'}},
            'extra-bindings': {'foo': {}},
            'peers': {'peer': {'interface': 'bar'}},
            'requires': {
                'rel': {'interface': 'bar'},
                'sub': {'interface': 'bar', 'scope': 'container'},
            },
            'resources': {'foo': {'type': 'bar'}},
            'storage': {'foo': {'type': 'file'}},
        },
        config={'options': {'foo': {'type': 'string'}}},
    )

    if attribute == 'config':
        # There's no State.get_config, we just get it directly.
        obj_out = state_out.config
    elif attribute == 'opened_ports':
        # There's no State.get_opened_ports, because in a charm tests you just
        # want to assert the port is/is not in the set.
        obj_out = next(p for p in state_out.opened_ports if p == obj_in)
    elif attribute == 'secrets':
        # State.get_secret only takes keyword arguments, while the others take
        # only positional arguments.
        obj_out = state_out.get_secret(id=obj_in.id)
    elif attribute == 'resources':
        # Charms can't change resources, so there's no State.get_resource.
        obj_out = next(r for r in state_out.resources if r == obj_in)
    else:
        obj_out = getattr(state_out, get_method)(getattr(obj_in, key_attr))
    assert obj_in is not obj_out


@pytest.mark.parametrize(
    'relation_type',
    [
        Relation,
        PeerRelation,
        SubordinateRelation,
    ],
)
def test_state_immutable_with_changed_data_relation(relation_type: type[RelationBase], mycharm):
    def event_handler(charm: CharmBase, _):
        rel = charm.model.get_relation(relation_type.__name__)
        assert rel is not None
        rel.data[charm.app]['a'] = 'b'
        rel.data[charm.unit]['c'] = 'd'

    mycharm._call = event_handler

    relation_in = relation_type(relation_type.__name__)

    state_in = State(relations={relation_in}, leader=True)

    state_out = trigger(
        state_in,
        event='start',
        charm_type=mycharm,
        meta={
            'name': 'foo',
            'peers': {'PeerRelation': {'interface': 'bar'}},
            'requires': {
                'Relation': {'interface': 'bar'},
                'SubordinateRelation': {'interface': 'bar', 'scope': 'container'},
            },
        },
    )

    relation_out = state_out.get_relation(relation_in.id)
    assert not relation_in.local_app_data
    assert relation_out.local_app_data == {'a': 'b'}
    assert relation_out.local_unit_data == {'c': 'd', **_DEFAULT_JUJU_DATABAG}


def test_state_immutable_with_changed_data_container(mycharm):
    layer_name = 'my-layer'
    layer = ops.pebble.Layer({
        'services': {
            'foo': {
                'command': 'bar',
                'override': 'replace',
            },
        }
    })

    def event_handler(charm: CharmBase, _):
        container = charm.model.unit.get_container('foo')
        container.add_layer(layer_name, layer, combine=True)

    mycharm._call = event_handler

    container_in = Container('foo', can_connect=True)
    state_in = State(containers={container_in})

    state_out = trigger(
        state_in,
        event='start',
        charm_type=mycharm,
        meta={
            'name': 'foo',
            'containers': {'foo': {'resource': 'bar'}},
        },
    )

    container_out = state_out.get_container(container_in.name)
    assert not container_in.layers
    assert container_out.layers == {layer_name: layer}


def test_state_immutable_with_changed_data_ports(mycharm):
    def event_handler(charm: CharmBase, _):
        charm.model.unit.open_port(protocol='tcp', port=80)

    mycharm._call = event_handler

    state_in = State()
    state_out = trigger(
        state_in,
        event='start',
        charm_type=mycharm,
        meta={'name': 'foo'},
    )

    assert not state_in.opened_ports
    assert state_out.opened_ports == {TCPPort(80)}


def test_state_immutable_with_changed_data_secret(mycharm):
    def event_handler(charm: CharmBase, _):
        secret = charm.model.get_secret(label='my-secret')
        secret.set_content({'password': 'bar'})

    mycharm._call = event_handler

    secret_in = Secret({'password': 'foo'}, label='my-secret', owner='unit')
    state_in = State(secrets={secret_in})

    state_out = trigger(
        state_in,
        event='start',
        charm_type=mycharm,
        meta={'name': 'foo'},
    )

    secret_out = state_out.get_secret(id=secret_in.id)
    assert secret_in.latest_content == {'password': 'foo'}
    assert secret_out.latest_content == {'password': 'bar'}


def test_state_immutable_with_changed_data_stored_state():
    class MyCharm(ops.CharmBase):
        _stored = ops.StoredState()

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            self._stored.set_default(seen=set())
            framework.observe(self.on.start, self._on_start)

        def _on_start(self, event: ops.StartEvent):
            self._stored.seen.add(str(event))

    stored_state_in = StoredState(owner_path='MyCharm')
    state_in = State(stored_states={stored_state_in})

    state_out = trigger(
        state_in,
        event='start',
        charm_type=MyCharm,
        meta={'name': 'foo'},
    )

    stored_state_out = state_out.get_stored_state(
        stored_state_in.name, owner_path=stored_state_in.owner_path
    )
    assert not stored_state_in.content
    assert 'seen' in stored_state_out.content


@pytest.mark.parametrize(
    'rockcraft',
    [
        pytest.param({'summary': 'Empty Layer'}, id='empty-layer'),
        pytest.param(
            {'summary': 'Layer with description', 'description': 'This is a test layer.'},
            id='layer-with-description',
        ),
        pytest.param(
            {
                'summary': 'Layer with a service',
                'services': {'srv1': {'override': 'replace', 'command': 'echo hello'}},
            },
            id='layer-with-service',
        ),
        pytest.param(
            {
                'summary': 'Layer with a check',
                'checks': {'chk1': {'override': 'merge', 'exec': {'command': 'echo hello'}}},
            },
            id='layer-with-check',
        ),
        # This is from the example in the Rockcraft documentation:
        # https://documentation.ubuntu.com/rockcraft/en/latest/reference/rockcraft.yaml/
        # (with an extra service and two checks added).
        pytest.param(
            {
                'name': 'hello',
                'title': 'Hello World',
                'summary': 'A Hello World rock',
                'description': 'This is an example of a Rockcraft project for a Hello World rock.',
                'version': 'latest',
                'base': 'bare',
                'build-base': 'ubuntu@22.04',
                'license': 'Apache-2.0',
                'run-user': '_daemon_',
                'environment': {
                    'FOO': 'bar',
                },
                'services': {
                    'hello': {
                        'override': 'replace',
                        'command': '/usr/bin/hello -t',
                        'environment': {
                            'VAR1': 'value',
                            'VAR2': 'other value',
                        },
                    },
                    'hello-echo': {
                        'override': 'replace',
                        'command': 'echo hello',
                    },
                },
                'checks': {
                    'chk1': {
                        'override': 'merge',
                        'exec': {
                            'command': '/usr/bin/hello -c',
                        },
                    },
                    'chk2': {
                        'override': 'merge',
                        'level': 'ready',
                        'tcp': {
                            'port': 8000,
                        },
                    },
                },
                'platforms': {
                    'amd64': {},
                    'armhf': {
                        'build-on': ['armhf', 'arm64'],
                    },
                    'ibm': {
                        'build-on': ['s390x'],
                        'build-for': 's390x',
                    },
                },
                'parts': {
                    'hello': {
                        'plugin': 'nil',
                        'stage-packages': ['hello'],
                    },
                },
            },
            id='rockcraft-doc-example',
        ),
    ],
)
def test_layer_from_rockcraft(rockcraft: dict[str, Any]):
    with tempfile.NamedTemporaryFile(mode='w', suffix='.yaml') as f:
        yaml.safe_dump(rockcraft, f)
        rockcraft_path = f.name
        layer = layer_from_rockcraft(rockcraft_path)
    assert layer.summary == rockcraft['summary']
    if 'description' in rockcraft:
        assert rockcraft['description'] in layer.description
        assert rockcraft_path in layer.description
    assert len(layer.services) == len(rockcraft.get('services', {}))
    for service_name, service in rockcraft.get('services', {}).items():
        assert service_name in layer.services
        layer_service = layer.services[service_name]
        assert layer_service.command == service['command']
        assert layer_service.override == service['override']
        if 'environment' in service:
            assert layer_service.environment == service['environment']
    assert len(layer.checks) == len(rockcraft.get('checks', {}))
    for check_name, check in rockcraft.get('checks', {}).items():
        assert check_name in layer.checks
        layer_check = layer.checks[check_name]
        assert layer_check.override == check['override']
        layer_check_level = layer_check.level
        if hasattr(layer_check_level, 'value'):
            assert isinstance(layer_check_level, ops.pebble.CheckLevel)
            layer_check_level = layer_check_level.value
        assert layer_check_level == check.get('level', ops.pebble.CheckLevel.UNSET.value)
        if 'exec' in check:
            assert layer_check.exec is not None and 'command' in check['exec']
            assert layer_check.exec['command'] == check['exec']['command']
        if 'tcp' in check:
            assert layer_check.tcp is not None and 'port' in check['tcp']
            assert layer_check.tcp['port'] == check['tcp']['port']


def test_layer_from_rockcraft_safe():
    dangerous_yaml = """
    !!python/object/apply:os.system ["echo unsafe"]
    """
    with tempfile.NamedTemporaryFile(mode='w', suffix='.yaml') as f:
        f.write(dangerous_yaml)
        f.flush()
        rockcraft_path = f.name
        with pytest.raises(yaml.YAMLError):
            layer_from_rockcraft(rockcraft_path)


def test_state_from_context():
    meta = {
        'name': 'sam',
        'containers': {'container': {}},
        'peers': {'peer': {'interface': 'friend'}},
        'requires': {
            'relreq': {'interface': 'across'},
            'sub': {'interface': 'below', 'scope': 'container'},
        },
        'provides': {'relpro': {'interface': 'around'}},
        'storage': {'storage': {}},
    }

    class Charm(ops.CharmBase):
        _stored = ops.StoredState()

    ctx = Context(
        Charm, meta=meta, config={'options': {'foo': {'type': 'string', 'default': 'bar'}}}
    )
    state = State.from_context(ctx)
    assert state.config == {'foo': 'bar'}
    assert isinstance(state.containers, frozenset)
    assert len(state.containers) == 1
    assert state.get_container('container').can_connect
    assert isinstance(state.relations, frozenset)
    assert len(state.relations) == 4
    assert state.get_relations('peer')[0].interface == 'friend'
    assert state.get_relations('relreq')[0].interface == 'across'
    assert state.get_relations('relpro')[0].interface == 'around'
    assert state.get_relations('sub')[0].interface == 'below'
    assert isinstance(state.storages, frozenset)
    assert len(state.storages) == 1
    assert next(iter(state.storages)).name == 'storage'
    assert isinstance(state.stored_states, frozenset)
    assert len(state.stored_states) == 1
    assert state.get_stored_state('_stored', owner_path='Charm').name == '_stored'


def test_state_from_context_extend():
    meta = {
        'name': 'sam',
        'containers': {'container': {}},
        'peers': {'peer': {'interface': 'friend'}},
        'requires': {
            'relreq': {'interface': 'across'},
            'sub': {'interface': 'below', 'scope': 'container'},
        },
        'provides': {'relpro': {'interface': 'around'}},
        'storage': {'storage': {}},
    }

    class Charm(ops.CharmBase):
        _stored = ops.StoredState()

    ctx = Context(
        Charm, meta=meta, config={'options': {'foo': {'type': 'string', 'default': 'bar'}}}
    )
    container = Container(name='other-container', can_connect=False)
    relation = Relation('db', remote_app_data={'a': 'b'})
    stored_state = StoredState(name='_stored', owner_path='Charm', content={'foo': 'bar'})
    state = State.from_context(
        ctx,
        config={'foo': 'baz'},
        leader=True,
        containers={container},
        relations={relation},
        stored_states={stored_state},
    )
    assert state.leader is True
    assert state.config == {'foo': 'baz'}
    assert isinstance(state.containers, frozenset)
    assert len(state.containers) == 2
    assert state.get_container('container').can_connect
    assert not state.get_container('other-container').can_connect
    assert isinstance(state.relations, frozenset)
    assert len(state.relations) == 5
    assert state.get_relations('peer')[0].interface == 'friend'
    assert state.get_relations('relreq')[0].interface == 'across'
    assert state.get_relations('relpro')[0].interface == 'around'
    assert state.get_relations('sub')[0].interface == 'below'
    assert state.get_relation(relation.id).remote_app_data == {'a': 'b'}
    assert isinstance(state.storages, frozenset)
    assert len(state.storages) == 1
    assert next(iter(state.storages)).name == 'storage'
    assert isinstance(state.stored_states, frozenset)
    assert len(state.stored_states) == 1
    assert state.get_stored_state('_stored', owner_path='Charm').name == '_stored'
    assert state.get_stored_state('_stored', owner_path='Charm').content == {'foo': 'bar'}


def test_state_from_context_merge_config():
    ctx = Context(
        ops.CharmBase,
        meta={'name': 'alex'},
        config={'options': {'foo': {'type': 'string', 'default': 'str'}}},
    )
    state = State.from_context(ctx, config={'bar': 'baz'})
    assert state.config == {'foo': 'str', 'bar': 'baz'}


@pytest.mark.parametrize(
    'rel_type,endpoint',
    [(Relation, 'relreq'), (PeerRelation, 'peer'), (SubordinateRelation, 'sub')],
)
def test_state_from_context_skip_exiting_relation(rel_type: type[RelationBase], endpoint: str):
    meta = {
        'name': 'sam',
        'peers': {'peer': {'interface': 'friend'}},
        'requires': {
            'relreq': {'interface': 'across'},
            'sub': {'interface': 'below', 'scope': 'container'},
        },
    }
    rel = rel_type(endpoint, local_app_data={'a': 'b'})
    ctx = Context(ops.CharmBase, meta=meta)
    state = State.from_context(ctx, relations={rel})
    # We should have the relation passed in, that has data, not an empty one
    # from the context.
    assert state.get_relation(rel.id).local_app_data == {'a': 'b'}


def test_state_from_context_skip_exiting_container():
    meta = {
        'name': 'sam',
        'containers': {'container': {}},
    }
    notice = Notice('example.com/notice')
    container = Container(name='container', notices=[notice])
    ctx = Context(ops.CharmBase, meta=meta)
    state = State.from_context(ctx, containers={container})
    # We should have the container passed in, that has notices and has the
    # default of can_connect=False, not the one from the context.
    assert state.get_container(container.name).notices == [notice]
    assert state.get_container(container.name).can_connect is False


def test_state_from_context_skip_exiting_storage():
    meta = {
        'name': 'sam',
        'storage': {'storage': {}},
    }
    next_index = _next_storage_index(update=False)
    storage = Storage(name='storage')
    assert storage.index == next_index
    next_index = _next_storage_index(update=False)
    ctx = Context(ops.CharmBase, meta=meta)
    state = State.from_context(ctx, storages={storage})
    # We should have the storage passed in. If the context created a new one,
    # the index would have increased.
    assert state.get_storage(storage.name, index=storage.index) == storage
    assert _next_storage_index(update=False) == next_index


def test_state_from_context_skip_exiting_stored_state():
    class Charm(ops.CharmBase):
        _stored = ops.StoredState()

    meta = {
        'name': 'sam',
    }
    stored_state = StoredState(name='_stored', owner_path='Charm', content={'foo': 'bar'})
    ctx = Context(Charm, meta=meta)
    state = State.from_context(ctx, stored_states={stored_state})
    # We should have the stored state passed in, that has content, not an empty
    # one from the context.
    assert state.get_stored_state(
        stored_state.name, owner_path=stored_state.owner_path
    ).content == {'foo': 'bar'}


def _make_generator(items: Iterable[Any]) -> Generator[Any]:
    return (item for item in items)


@pytest.mark.parametrize('iterable', [frozenset, tuple, list, _make_generator])
def test_state_from_non_sets(iterable: Callable[..., Any]):
    meta = {
        'name': 'sam',
        'containers': {'container': {}},
        'peers': {'peer': {'interface': 'friend'}},
        'requires': {
            'relreq': {'interface': 'across'},
            'sub': {'interface': 'below', 'scope': 'container'},
        },
        'provides': {'relpro': {'interface': 'around'}},
        'storage': {'storage': {}},
    }

    class Charm(ops.CharmBase):
        _stored = ops.StoredState()

    ctx = Context(Charm, meta=meta)
    container = Container(name='other-container', can_connect=False)
    relation = Relation('db', remote_app_data={'a': 'b'})
    stored_state = StoredState(name='_stored', owner_path='Charm', content={'foo': 'bar'})
    state = State.from_context(
        ctx,
        containers=iterable({container}),
        relations=iterable({relation}),
        stored_states=iterable({stored_state}),
    )
    assert isinstance(state.containers, frozenset)
    assert len(state.containers) == 2
    assert isinstance(state.relations, frozenset)
    assert len(state.relations) == 5
    assert isinstance(state.storages, frozenset)
    assert len(state.storages) == 1
    assert isinstance(state.stored_states, frozenset)
    assert len(state.stored_states) == 1



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import pytest
from scenario import Context
from scenario.state import State, _Action, _next_action_id

from ops._private.harness import ActionFailed
from ops.charm import ActionEvent, CharmBase
from ops.framework import Framework


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(CharmBase):
        _evt_handler = None

        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if handler := self._evt_handler:
                handler(event)

    return MyCharm


@pytest.mark.parametrize('baz_value', (True, False))
def test_action_event(mycharm, baz_value):
    ctx = Context(
        mycharm,
        meta={'name': 'foo'},
        actions={'foo': {'params': {'bar': {'type': 'number'}, 'baz': {'type': 'boolean'}}}},
    )
    state = ctx.run(ctx.on.action('foo', params={'baz': baz_value, 'bar': 10}), State())

    assert isinstance(state, State)
    evt = ctx.emitted_events[0]
    assert isinstance(evt, ActionEvent)
    assert evt.params['bar'] == 10
    assert evt.params['baz'] is baz_value


def test_action_no_results():
    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.act_action, self._on_act_action)

        def _on_act_action(self, _):
            pass

    ctx = Context(MyCharm, meta={'name': 'foo'}, actions={'act': {}})
    ctx.run(ctx.on.action('act'), State())
    assert ctx.action_results is None
    assert ctx.action_logs == []


@pytest.mark.parametrize('res_value', ('one', 1, [2], ['bar'], (1,), {1, 2}))
def test_action_event_results_invalid(mycharm, res_value):
    def handle_evt(charm: CharmBase, evt: ActionEvent):
        with pytest.raises((TypeError, AttributeError)):
            evt.set_results(res_value)

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={'name': 'foo'}, actions={'foo': {}})
    ctx.run(ctx.on.action('foo'), State())


@pytest.mark.parametrize('res_value', ({'a': {'b': {'c'}}}, {'d': 'e'}))
def test_action_event_results_valid(mycharm, res_value):
    def handle_evt(_: CharmBase, evt):
        if not isinstance(evt, ActionEvent):
            return
        evt.set_results(res_value)
        evt.log('foo')
        evt.log('bar')

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={'name': 'foo'}, actions={'foo': {}})

    ctx.run(ctx.on.action('foo'), State())

    assert ctx.action_results == res_value


@pytest.mark.parametrize('res_value', ({'a': {'b': {'c'}}}, {'d': 'e'}))
def test_action_event_outputs(mycharm, res_value):
    def handle_evt(_: CharmBase, evt: ActionEvent):
        if not isinstance(evt, ActionEvent):
            return

        evt.set_results({'my-res': res_value})
        evt.log('log1')
        evt.log('log2')
        evt.fail('failed becozz')

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={'name': 'foo'}, actions={'foo': {}})
    with pytest.raises(ActionFailed) as exc_info:
        ctx.run(ctx.on.action('foo'), State())
    assert exc_info.value.message == 'failed becozz'
    assert ctx.action_results == {'my-res': res_value}
    assert ctx.action_logs == ['log1', 'log2']


def test_action_event_fail(mycharm):
    def handle_evt(_: CharmBase, evt: ActionEvent):
        if not isinstance(evt, ActionEvent):
            return
        evt.fail('action failed!')

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={'name': 'foo'}, actions={'foo': {}})
    with pytest.raises(ActionFailed) as exc_info:
        ctx.run(ctx.on.action('foo'), State())
    assert exc_info.value.message == 'action failed!'


def test_action_event_fail_context_manager(mycharm):
    def handle_evt(_: CharmBase, evt: ActionEvent):
        if not isinstance(evt, ActionEvent):
            return
        evt.fail('action failed!')

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={'name': 'foo'}, actions={'foo': {}})
    with pytest.raises(ActionFailed) as exc_info:
        with ctx(ctx.on.action('foo'), State()):
            assert False, 'ActionFailed should be raised in the context manager.'
    assert exc_info.value.message == 'action failed!'


def test_action_continues_after_fail():
    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.foo_action, self._on_foo_action)

        def _on_foo_action(self, event):
            event.log('starting')
            event.set_results({'initial': 'result'})
            event.fail('oh no!')
            event.set_results({'final': 'result'})

    ctx = Context(MyCharm, meta={'name': 'foo'}, actions={'foo': {}})
    with pytest.raises(ActionFailed) as exc_info:
        ctx.run(ctx.on.action('foo'), State())
    assert exc_info.value.message == 'oh no!'
    assert ctx.action_logs == ['starting']
    assert ctx.action_results == {'initial': 'result', 'final': 'result'}


def test_action_event_has_id(mycharm):
    def handle_evt(_: CharmBase, evt: ActionEvent):
        if not isinstance(evt, ActionEvent):
            return
        assert isinstance(evt.id, str) and evt.id != ''

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={'name': 'foo'}, actions={'foo': {}})
    ctx.run(ctx.on.action('foo'), State())


def test_action_event_has_override_id(mycharm):
    uuid = '0ddba11-cafe-ba1d-5a1e-dec0debad'

    def handle_evt(charm: CharmBase, evt: ActionEvent):
        if not isinstance(evt, ActionEvent):
            return
        assert evt.id == uuid

    mycharm._evt_handler = handle_evt

    ctx = Context(mycharm, meta={'name': 'foo'}, actions={'foo': {}})
    ctx.run(ctx.on.action('foo', id=uuid), State())


def test_two_actions_same_context():
    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            framework.observe(self.on.foo_action, self._on_foo_action)
            framework.observe(self.on.bar_action, self._on_bar_action)

        def _on_foo_action(self, event):
            event.log('foo')
            event.set_results({'foo': 'result'})

        def _on_bar_action(self, event):
            event.log('bar')
            event.set_results({'bar': 'result'})

    ctx = Context(MyCharm, meta={'name': 'foo'}, actions={'foo': {}, 'bar': {}})
    ctx.run(ctx.on.action('foo'), State())
    assert ctx.action_results == {'foo': 'result'}
    assert ctx.action_logs == ['foo']
    # Not recommended, but run another action in the same context.
    ctx.run(ctx.on.action('bar'), State())
    assert ctx.action_results == {'bar': 'result'}
    assert ctx.action_logs == ['bar']


def test_positional_arguments():
    with pytest.raises(TypeError):
        _Action('foo', {})


def test_default_arguments():
    expected_id = _next_action_id(update=False)
    name = 'foo'
    action = _Action(name)
    assert action.name == name
    assert action.params == {}
    assert action.id == expected_id



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

from typing import ClassVar

import pytest
from scenario.state import State, _CharmSpec, _Event

from ops.charm import CharmBase, CharmEvents
from ops.framework import EventBase, EventSource, Framework, Object

from ..helpers import trigger


class QuxEvent(EventBase):
    pass


class SubEvent(EventBase):
    pass


@pytest.fixture(scope='function')
def mycharm():
    class MyCharmEvents(CharmEvents):
        qux = EventSource(QuxEvent)

    class MySubEvents(CharmEvents):
        sub = EventSource(SubEvent)

    class Sub(Object):
        on = MySubEvents()

    class MyCharm(CharmBase):
        on = MyCharmEvents()
        evts: ClassVar[list[EventBase]] = []

        def __init__(self, framework: Framework):
            super().__init__(framework)
            self.sub = Sub(self, 'sub')
            self.framework.observe(self.sub.on.sub, self._on_event)
            self.framework.observe(self.on.qux, self._on_event)

        def _on_event(self, e):
            MyCharm.evts.append(e)

    return MyCharm


@pytest.mark.parametrize('evt_name', ('rubbish', 'foo', 'bar'))
def test_rubbish_event_raises(mycharm: CharmBase, evt_name: str):
    with pytest.raises(AttributeError):
        trigger(State(), evt_name, mycharm, meta={'name': 'foo'})


def test_rubbish_pebble_ready_event_raises(mycharm: CharmBase, monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv('SCENARIO_SKIP_CONSISTENCY_CHECKS', '1')
    # else it will whine about the container not being in state and meta;
    # but if we put the container in meta, it will actually register an event!
    with pytest.raises(AttributeError):
        trigger(State(), 'kazoo_pebble_ready', mycharm, meta={'name': 'foo'})


@pytest.mark.parametrize('evt_name', ('qux',))
def test_custom_events_fail(mycharm, evt_name):
    with pytest.raises(AttributeError):
        trigger(State(), evt_name, mycharm, meta={'name': 'foo'})


# cfr: https://github.com/PietroPasotti/ops-scenario/pull/11#discussion_r1101694961
@pytest.mark.parametrize('evt_name', ('sub',))
def test_custom_events_sub_raise(mycharm, evt_name):
    with pytest.raises(AttributeError):
        trigger(State(), evt_name, mycharm, meta={'name': 'foo'})


@pytest.mark.parametrize(
    'evt_name, expected',
    (
        ('qux', False),
        ('sub', False),
        ('start', True),
        ('install', True),
        ('config-changed', True),
        ('foo-relation-changed', True),
        ('bar-relation-changed', True),
    ),
)
def test_is_custom_event(mycharm, evt_name, expected):
    spec = _CharmSpec(charm_type=mycharm, meta={'name': 'mycharm', 'requires': {'foo': {}}})
    assert _Event(evt_name)._is_builtin_event(spec) is expected



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import pytest
import scenario

import ops


class MyCharm(ops.CharmBase):
    def __init__(self, framework: ops.Framework):
        super().__init__(framework)
        for evt in self.on.events().values():
            self.framework.observe(evt, self._on_event)

    def _on_event(self, event):
        pass


def test_get_cloud_spec():
    scenario_cloud_spec = scenario.CloudSpec(
        type='lxd',
        name='localhost',
        endpoint='https://127.0.0.1:8443',
        credential=scenario.CloudCredential(
            auth_type='clientcertificate',
            attributes={
                'client-cert': 'foo',
                'client-key': 'bar',
                'server-cert': 'baz',
            },
        ),
    )
    expected_cloud_spec = ops.CloudSpec(
        type='lxd',
        name='localhost',
        endpoint='https://127.0.0.1:8443',
        credential=ops.CloudCredential(
            auth_type='clientcertificate',
            attributes={
                'client-cert': 'foo',
                'client-key': 'bar',
                'server-cert': 'baz',
            },
        ),
    )
    ctx = scenario.Context(MyCharm, meta={'name': 'foo'}, app_trusted=True)
    state = scenario.State(
        model=scenario.Model(name='lxd-model', type='lxd', cloud_spec=scenario_cloud_spec),
    )
    with ctx(ctx.on.start(), state=state) as mgr:
        assert mgr.charm.model.get_cloud_spec() == expected_cloud_spec


def test_get_cloud_spec_error():
    ctx = scenario.Context(MyCharm, meta={'name': 'foo'})
    state = scenario.State(model=scenario.Model(name='lxd-model', type='lxd'))
    with ctx(ctx.on.start(), state) as mgr:
        with pytest.raises(ops.ModelError):
            mgr.charm.model.get_cloud_spec()


def test_get_cloud_spec_untrusted():
    cloud_spec = scenario.CloudSpec(type='lxd', name='localhost')
    ctx = scenario.Context(MyCharm, meta={'name': 'foo'})
    state = scenario.State(
        model=scenario.Model(name='lxd-model', type='lxd', cloud_spec=cloud_spec),
    )
    with ctx(ctx.on.start(), state) as mgr:
        with pytest.raises(ops.ModelError):
            mgr.charm.model.get_cloud_spec()



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

from collections.abc import Mapping
from typing import Any, ClassVar

import pytest
from scenario.state import State, StoredState

import ops
from ops.framework import StoredState as OpsStoredstate
from tests.helpers import trigger


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(ops.CharmBase):
        META: Mapping[str, Any] = {'name': 'mycharm'}

        _read: ClassVar[dict[str, Any]] = {}
        _stored = OpsStoredstate()
        _stored2 = OpsStoredstate()

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            self._stored.set_default(foo='bar', baz={12: 142})
            self._stored2.set_default(foo='bar', baz={12: 142})
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, _: ops.EventBase):
            self._read['foo'] = self._stored.foo
            self._read['baz'] = self._stored.baz

    return MyCharm


def test_stored_state_default(mycharm):
    out = trigger(State(), 'start', mycharm, meta=mycharm.META)
    assert out.get_stored_state('_stored', owner_path='MyCharm').content == {
        'foo': 'bar',
        'baz': {12: 142},
    }
    assert out.get_stored_state('_stored2', owner_path='MyCharm').content == {
        'foo': 'bar',
        'baz': {12: 142},
    }


def test_stored_state_initialized(mycharm):
    out = trigger(
        State(
            stored_states={
                StoredState(owner_path='MyCharm', name='_stored', content={'foo': 'FOOX'}),
            }
        ),
        'start',
        mycharm,
        meta=mycharm.META,
    )
    assert out.get_stored_state('_stored', owner_path='MyCharm').content == {
        'foo': 'FOOX',
        'baz': {12: 142},
    }
    assert out.get_stored_state('_stored2', owner_path='MyCharm').content == {
        'foo': 'bar',
        'baz': {12: 142},
    }


def test_positional_arguments():
    with pytest.raises(TypeError):
        StoredState('_stored', '')


def test_default_arguments():
    s = StoredState()
    assert s.name == '_stored'
    assert s.owner_path is None
    assert s.content == {}
    assert s._data_type_name == 'StoredStateData'



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import collections
import datetime
from typing import Literal, cast
from unittest.mock import ANY

import pytest
from scenario import Context
from scenario.state import Relation, Secret, State

from ops.charm import CharmBase
from ops.framework import Framework
from ops.model import ModelError, SecretNotFoundError, SecretRotate
from ops.model import Secret as ops_Secret
from test.charms.test_secrets.src.charm import Result, SecretsCharm
from tests.helpers import trigger


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            pass

    return MyCharm


def test_get_secret_no_secret(mycharm):
    ctx = Context(mycharm, meta={'name': 'local'})
    with ctx(ctx.on.update_status(), State()) as mgr:
        with pytest.raises(SecretNotFoundError):
            assert mgr.charm.model.get_secret(id='foo')
        with pytest.raises(SecretNotFoundError):
            assert mgr.charm.model.get_secret(label='foo')


@pytest.mark.parametrize('owner', ('app', 'unit'))
def test_get_secret(mycharm, owner):
    ctx = Context(mycharm, meta={'name': 'local'})
    secret = Secret({'a': 'b'}, owner=owner)
    with ctx(
        state=State(secrets={secret}),
        event=ctx.on.update_status(),
    ) as mgr:
        assert mgr.charm.model.get_secret(id=secret.id).get_content()['a'] == 'b'


@pytest.mark.parametrize('owner', ('app', 'unit'))
def test_get_secret_get_refresh(mycharm, owner):
    ctx = Context(mycharm, meta={'name': 'local'})
    secret = Secret(
        tracked_content={'a': 'b'},
        latest_content={'a': 'c'},
        owner=owner,
    )
    with ctx(
        ctx.on.update_status(),
        State(secrets={secret}),
    ) as mgr:
        charm = mgr.charm
        assert charm.model.get_secret(id=secret.id).get_content(refresh=True)['a'] == 'c'


@pytest.mark.parametrize('app', (True, False))
def test_get_secret_nonowner_peek_update(mycharm, app):
    ctx = Context(mycharm, meta={'name': 'local'})
    secret = Secret(
        tracked_content={'a': 'b'},
        latest_content={'a': 'c'},
    )
    with ctx(
        ctx.on.update_status(),
        State(
            leader=app,
            secrets={secret},
        ),
    ) as mgr:
        charm = mgr.charm
        assert charm.model.get_secret(id=secret.id).get_content()['a'] == 'b'
        assert charm.model.get_secret(id=secret.id).peek_content()['a'] == 'c'
        # Verify that the peek has not refreshed:
        assert charm.model.get_secret(id=secret.id).get_content()['a'] == 'b'

        assert charm.model.get_secret(id=secret.id).get_content(refresh=True)['a'] == 'c'
        assert charm.model.get_secret(id=secret.id).get_content()['a'] == 'c'


@pytest.mark.parametrize('owner', ('app', 'unit'))
def test_get_secret_owner_peek_update(mycharm, owner):
    ctx = Context(mycharm, meta={'name': 'local'})
    secret = Secret(
        tracked_content={'a': 'b'},
        latest_content={'a': 'c'},
        owner=owner,
    )
    with ctx(
        ctx.on.update_status(),
        State(
            secrets={secret},
        ),
    ) as mgr:
        charm = mgr.charm
        assert charm.model.get_secret(id=secret.id).get_content()['a'] == 'b'
        assert charm.model.get_secret(id=secret.id).peek_content()['a'] == 'c'
        # Verify that the peek has not refreshed:
        assert charm.model.get_secret(id=secret.id).get_content()['a'] == 'b'
        assert charm.model.get_secret(id=secret.id).get_content(refresh=True)['a'] == 'c'


@pytest.mark.parametrize('owner', ('app', 'unit'))
def test_secret_changed_owner_evt_fails(mycharm, owner):
    ctx = Context(mycharm, meta={'name': 'local'})
    secret = Secret(
        tracked_content={'a': 'b'},
        latest_content={'a': 'c'},
        owner=owner,
    )
    with pytest.raises(ValueError):
        _ = ctx.on.secret_changed(secret)


@pytest.mark.parametrize(
    'evt_suffix,revision',
    [
        ('rotate', None),
        ('expired', 1),
        ('remove', 1),
    ],
)
def test_consumer_events_failures(mycharm, evt_suffix, revision):
    ctx = Context(mycharm, meta={'name': 'local'})
    secret = Secret(
        tracked_content={'a': 'b'},
        latest_content={'a': 'c'},
    )
    kwargs = {'secret': secret}
    if revision is not None:
        kwargs['revision'] = revision
    with pytest.raises(ValueError):
        _ = getattr(ctx.on, f'secret_{evt_suffix}')(**kwargs)


@pytest.mark.parametrize('app', (True, False))
def test_add(mycharm, app):
    ctx = Context(mycharm, meta={'name': 'local'})
    with ctx(
        ctx.on.update_status(),
        State(leader=app),
    ) as mgr:
        charm = mgr.charm
        if app:
            charm.app.add_secret({'foo': 'bar'}, label='mylabel')
        else:
            charm.unit.add_secret({'foo': 'bar'}, label='mylabel')
        output = mgr.run()

    assert output.secrets
    secret = output.get_secret(label='mylabel')
    assert secret.latest_content == secret.tracked_content == {'foo': 'bar'}
    assert secret.label == 'mylabel'


def test_set_legacy_behaviour(mycharm):
    # in juju < 3.1.7, secret owners always used to track the latest revision.
    # ref: https://bugs.launchpad.net/juju/+bug/2037120
    ctx = Context(mycharm, meta={'name': 'local'}, juju_version='3.1.6')
    rev1, rev2 = {'foo': 'bar'}, {'foo': 'baz', 'qux': 'roz'}
    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        charm = mgr.charm
        secret: ops_Secret = charm.unit.add_secret(rev1, label='mylabel')
        assert (
            secret.get_content()
            == secret.peek_content()
            == secret.get_content(refresh=True)
            == rev1
        )

        secret.set_content(rev2)
        # We need to get the secret again, because ops caches the content in
        # the object.
        secret: ops_Secret = charm.model.get_secret(label='mylabel')
        assert (
            secret.get_content()
            == secret.peek_content()
            == secret.get_content(refresh=True)
            == rev2
        )

        state_out = mgr.run()

    assert (
        state_out.get_secret(label='mylabel').tracked_content
        == state_out.get_secret(label='mylabel').latest_content
        == rev2
    )


def test_set(mycharm):
    ctx = Context(mycharm, meta={'name': 'local'})
    rev1, rev2 = {'foo': 'bar'}, {'foo': 'baz', 'qux': 'roz'}
    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        charm = mgr.charm
        secret: ops_Secret = charm.unit.add_secret(rev1, label='mylabel')
        assert (
            secret.get_content()
            == secret.peek_content()
            == secret.get_content(refresh=True)
            == rev1
        )

        # TODO: if this is done in the same event hook, it's more complicated
        # than this. Figure out what we should do here.
        # Also the next test, for Juju 3.3
        secret.set_content(rev2)
        assert secret.get_content() == rev1
        assert secret.peek_content() == secret.get_content(refresh=True) == rev2

        state_out = mgr.run()

    assert (
        state_out.get_secret(label='mylabel').tracked_content
        == state_out.get_secret(label='mylabel').latest_content
        == rev2
    )


def test_set_juju33(mycharm):
    ctx = Context(mycharm, meta={'name': 'local'}, juju_version='3.3.1')
    rev1, rev2 = {'foo': 'bar'}, {'foo': 'baz', 'qux': 'roz'}
    with ctx(
        ctx.on.update_status(),
        State(),
    ) as mgr:
        charm = mgr.charm
        secret: ops_Secret = charm.unit.add_secret(rev1, label='mylabel')
        assert secret.get_content() == rev1

        secret.set_content(rev2)
        assert secret.get_content() == rev1
        assert secret.peek_content() == rev2
        assert secret.get_content(refresh=True) == rev2

        state_out = mgr.run()

    assert (
        state_out.get_secret(label='mylabel').tracked_content
        == state_out.get_secret(label='mylabel').latest_content
        == rev2
    )


@pytest.mark.parametrize('app', (True, False))
def test_meta(mycharm, app):
    ctx = Context(mycharm, meta={'name': 'local'})
    secret = Secret(
        {'a': 'b'},
        owner='app' if app else 'unit',
        label='mylabel',
        description='foobarbaz',
        rotate=SecretRotate.HOURLY,
    )
    with ctx(
        ctx.on.update_status(),
        State(
            leader=True,
            secrets={secret},
        ),
    ) as mgr:
        charm = mgr.charm
        assert charm.model.get_secret(label='mylabel')

        secret = charm.model.get_secret(id=secret.id)
        info = secret.get_info()

        assert secret.label is None
        assert info.description == 'foobarbaz'
        assert info.label == 'mylabel'
        assert info.rotation == SecretRotate.HOURLY


@pytest.mark.parametrize('leader', (True, False))
@pytest.mark.parametrize('owner', ('app', 'unit', None))
def test_secret_permission_model(mycharm, leader, owner):
    expect_manage = bool(
        # if you're the leader and own this app secret
        (owner == 'app' and leader)
        # you own this secret
        or (owner == 'unit')
    )

    ctx = Context(mycharm, meta={'name': 'local'})
    secret = Secret(
        {'a': 'b'},
        label='mylabel',
        owner=owner,
        description='foobarbaz',
        rotate=SecretRotate.HOURLY,
    )
    secret_id = secret.id
    with ctx(
        ctx.on.update_status(),
        State(
            leader=leader,
            secrets={secret},
        ),
    ) as mgr:
        # can always view
        secret: ops_Secret = mgr.charm.model.get_secret(id=secret_id)
        assert secret.get_content()['a'] == 'b'
        assert secret.peek_content()
        assert secret.get_content(refresh=True)

        if expect_manage:
            assert secret.get_content()
            assert secret.peek_content()
            assert secret.get_content(refresh=True)

            assert secret.get_info()
            secret.set_content({'foo': 'boo'})
            assert secret.get_content() == {'a': 'b'}  # rev1!
            assert secret.get_content(refresh=True) == {'foo': 'boo'}

            secret.remove_all_revisions()

        else:  # cannot manage
            # nothing else to do directly if you can't get a hold of the Secret instance
            # but we can try some raw backend calls
            with pytest.raises(ModelError):
                secret.get_info()

            with pytest.raises(ModelError):
                secret.set_content(content={'boo': 'foo'})


@pytest.mark.parametrize('app', (True, False))
def test_grant(mycharm, app):
    ctx = Context(mycharm, meta={'name': 'local', 'requires': {'foo': {'interface': 'bar'}}})
    secret = Secret(
        {'a': 'b'},
        owner='unit',
        label='mylabel',
        description='foobarbaz',
        rotate=SecretRotate.HOURLY,
    )
    with ctx(
        ctx.on.update_status(),
        State(
            relations=[Relation('foo', 'remote')],
            secrets={secret},
        ),
    ) as mgr:
        charm = mgr.charm
        secret = charm.model.get_secret(label='mylabel')
        foo = charm.model.get_relation('foo')
        if app:
            secret.grant(relation=foo)
        else:
            secret.grant(relation=foo, unit=foo.units.pop())
        output = mgr.run()
    vals = list(output.get_secret(label='mylabel').remote_grants.values())
    assert vals == [{'remote'}] if app else [{'remote/0'}]


def test_update_metadata(mycharm):
    exp = datetime.datetime(2050, 12, 12)

    ctx = Context(mycharm, meta={'name': 'local'})
    secret = Secret(
        {'a': 'b'},
        owner='unit',
        label='mylabel',
    )
    with ctx(
        ctx.on.update_status(),
        State(
            secrets={secret},
        ),
    ) as mgr:
        secret = mgr.charm.model.get_secret(label='mylabel')
        secret.set_info(
            label='babbuccia',
            description='blu',
            expire=exp,
            rotate=SecretRotate.DAILY,
        )
        output = mgr.run()

    secret_out = output.get_secret(label='babbuccia')
    assert secret_out.label == 'babbuccia'
    assert secret_out.rotate == SecretRotate.DAILY
    assert secret_out.description == 'blu'
    assert secret_out.expire == exp


@pytest.mark.parametrize('leader', (True, False))
def test_grant_after_add(leader):
    class GrantingCharm(CharmBase):
        def __init__(self, *args):
            super().__init__(*args)
            self.framework.observe(self.on.start, self._on_start)

        def _on_start(self, _):
            if leader:
                secret = self.app.add_secret({'foo': 'bar'})
            else:
                secret = self.unit.add_secret({'foo': 'bar'})
            secret.grant(self.model.relations['bar'][0])

    state = State(leader=leader, relations={Relation('bar')})
    ctx = Context(GrantingCharm, meta={'name': 'foo', 'provides': {'bar': {'interface': 'bar'}}})
    ctx.run(ctx.on.start(), state)


def test_grant_nonowner(mycharm):
    secret = Secret(
        {'a': 'b'},
        label='mylabel',
        description='foobarbaz',
        rotate=SecretRotate.HOURLY,
    )
    secret_id = secret.id

    def post_event(charm: CharmBase):
        secret = charm.model.get_secret(id=secret_id)
        secret = charm.model.get_secret(label='mylabel')
        foo = charm.model.get_relation('foo')
        assert foo is not None

        with pytest.raises(ModelError):
            secret.grant(relation=foo)

    trigger(
        State(
            relations={Relation('foo', 'remote')},
            secrets={secret},
        ),
        'update_status',
        mycharm,
        meta={'name': 'local', 'requires': {'foo': {'interface': 'bar'}}},
        post_event=post_event,
    )


def test_add_grant_revoke_remove():
    class GrantingCharm(CharmBase):
        pass

    ctx = Context(GrantingCharm, meta={'name': 'foo', 'provides': {'bar': {'interface': 'bar'}}})
    relation_remote_app = 'remote_secret_desirerer'
    relation_id = 42

    state = State(
        leader=True,
        relations={Relation('bar', remote_app_name=relation_remote_app, id=relation_id)},
    )

    with ctx(ctx.on.start(), state) as mgr:
        charm = mgr.charm
        secret = charm.app.add_secret({'foo': 'bar'}, label='mylabel')
        bar_relation = charm.model.relations['bar'][0]

        secret.grant(bar_relation)
        output = mgr.run()

    assert output.secrets
    scenario_secret = output.get_secret(label='mylabel')
    assert relation_remote_app in scenario_secret.remote_grants[relation_id]

    with ctx(ctx.on.start(), output) as mgr:
        charm: GrantingCharm = mgr.charm
        secret = charm.model.get_secret(label='mylabel')
        secret.revoke(bar_relation)
        output = mgr.run()

    scenario_secret = output.get_secret(label='mylabel')
    assert scenario_secret.remote_grants == {}

    with ctx(ctx.on.start(), output) as mgr:
        charm: GrantingCharm = mgr.charm
        secret = charm.model.get_secret(label='mylabel')
        secret.remove_all_revisions()
        output = mgr.run()

    with pytest.raises(KeyError):
        output.get_secret(label='mylabel')


def test_secret_removed_event():
    class SecretCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.secret_remove, self._on_secret_remove)

        def _on_secret_remove(self, event):
            event.secret.remove_revision(event.revision)

    ctx = Context(SecretCharm, meta={'name': 'foo'})
    secret = Secret({'a': 'b'}, owner='app')
    old_revision = 42
    state = ctx.run(
        ctx.on.secret_remove(secret, revision=old_revision),
        State(leader=True, secrets={secret}),
    )
    assert secret in state.secrets
    assert ctx.removed_secret_revisions == [old_revision]


def test_secret_expired_event():
    class SecretCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.secret_expired, self._on_secret_expired)

        def _on_secret_expired(self, event):
            event.secret.set_content({'password': 'newpass'})
            event.secret.remove_revision(event.revision)

    ctx = Context(SecretCharm, meta={'name': 'foo'})
    secret = Secret({'password': 'oldpass'}, owner='app')
    old_revision = 42
    state = ctx.run(
        ctx.on.secret_expired(secret, revision=old_revision),
        State(leader=True, secrets={secret}),
    )
    assert state.get_secret(id=secret.id).latest_content == {'password': 'newpass'}
    assert ctx.removed_secret_revisions == [old_revision]


def test_remove_bad_revision():
    class SecretCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.secret_remove, self._on_secret_remove)

        def _on_secret_remove(self, event):
            with pytest.raises(ValueError):
                event.secret.remove_revision(event.revision)

    ctx = Context(SecretCharm, meta={'name': 'foo'})
    secret = Secret({'a': 'b'}, owner='app')
    ctx.run(
        ctx.on.secret_remove(secret, revision=secret._latest_revision),
        State(leader=True, secrets={secret}),
    )
    ctx.run(
        ctx.on.secret_remove(secret, revision=secret._tracked_revision),
        State(leader=True, secrets={secret}),
    )


def test_set_label_on_get():
    class SecretCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.start, self._on_start)

        def _on_start(self, _):
            id = self.unit.add_secret({'foo': 'bar'}).id
            secret = self.model.get_secret(id=id, label='label1')
            assert secret.label == 'label1'
            secret = self.model.get_secret(id=id, label='label2')
            assert secret.label == 'label2'

    ctx = Context(SecretCharm, meta={'name': 'foo'})
    state = ctx.run(ctx.on.start(), State())
    assert state.get_secret(label='label2').tracked_content == {'foo': 'bar'}


def test_no_additional_positional_arguments():
    with pytest.raises(TypeError):
        Secret({}, {})


def test_default_values():
    contents = {'foo': 'bar'}
    secret = Secret(contents)
    assert secret.latest_content == secret.tracked_content == contents
    assert secret.id.startswith('secret:')
    assert secret.label is None
    assert secret.description is None
    assert secret.owner is None
    assert secret.rotate is None
    assert secret.expire is None
    assert secret.remote_grants == {}


def test_add_secret(secrets_context: Context[SecretsCharm]):
    state = State(leader=True)
    state = secrets_context.run(secrets_context.on.action('add-secret'), state)

    result = cast('Result', secrets_context.action_results)
    assert result is not None
    assert result['secretid']

    scenario_secret = next(iter(state.secrets))

    common_assertions(scenario_secret, result)

    assert result == {
        'after': {
            'info': ANY,  # relying on scaffolding check
            'tracked': {'foo': 'bar'},
            'latest': {'foo': 'bar'},
        },
        'secretid': ANY,
    }

    assert scenario_secret._tracked_revision == scenario_secret._latest_revision


@pytest.mark.parametrize(
    'fields',
    [
        '',
        'label',
        'description',
        'expire',
        'rotate',
        'label,description',
        'description,expire,rotate',
        'label,description,expire,rotate',
    ],
)
def test_add_secret_with_metadata(secrets_context: Context[SecretsCharm], fields: str):
    state = State(leader=True)
    state = secrets_context.run(
        secrets_context.on.action('add-with-meta', params={'fields': fields}), state
    )
    scenario_secret = next(iter(state.secrets))
    result = cast('Result', secrets_context.action_results)
    assert 'after' in result
    assert result['after']
    info = result['after']['info']
    assert info

    common_assertions(scenario_secret, result)

    if 'label' in fields:
        assert scenario_secret.label == 'label1'
        assert info['label'] == 'label1'
    if 'description' in fields:
        assert scenario_secret.description == 'description1'
        assert info['description'] == 'description1'
    if 'expire' in fields:
        assert scenario_secret.expire == datetime.datetime(2020, 1, 1, 0, 0, 0)
        assert info['expires'] == datetime.datetime(2020, 1, 1, 0, 0, 0)
    if 'rotate' in fields:
        assert scenario_secret.rotate == SecretRotate.DAILY
        assert info['rotation'] == SecretRotate.DAILY
        # https://github.com/canonical/operator/issues/2104
        assert info['rotates'] is None

    assert scenario_secret._tracked_revision == scenario_secret._latest_revision


@pytest.mark.parametrize('lookup_by', ['id', 'label'])
@pytest.mark.parametrize(
    'flow',
    [
        'content,label,description,expire,rotate',
        'content,description,content,description',
        'rotate,content,rotate,content,rotate',
        'label,content,label,content',
    ],
)
def test_set_secret(
    secrets_context: Context[SecretsCharm], flow: str, lookup_by: Literal['id', 'label']
):
    secret = Secret({'some': 'content'}, owner='app', id='theid', label='thelabel')
    state = State(leader=True, secrets={secret})
    params = {'flow': flow, f'secret{lookup_by}': f'the{lookup_by}'}
    state = secrets_context.run(secrets_context.on.action('set-secret-flow', params=params), state)
    scenario_secret = state.get_secret(id='theid')
    result = cast('Result', secrets_context.action_results)
    assert 'after' in result
    assert result['after']
    info = result['after']['info']
    assert info

    common_assertions(scenario_secret, result)

    counts = collections.Counter(flow.split(','))
    if counts['content']:
        assert result['after']['latest'] == {'val': str(counts['content'])}
    if counts['label']:
        assert info['label'] == f'label{counts["label"]}'
    if counts['description']:
        assert info['description'] == f'description{counts["description"]}'
    if counts['expire']:
        assert info['expires'] == datetime.datetime(2010 + counts['expire'], 1, 1, 0, 0)
    if counts['rotate']:
        rotation_values = ['sentinel', *SecretRotate.__members__.values()]
        assert info['rotation'] == rotation_values[counts['rotate']]


def common_assertions(scenario_secret: Secret | None, result: Result):
    if scenario_secret:
        assert scenario_secret.owner == 'app'
        assert not scenario_secret.remote_grants

        assert result.get('after')
        info = result['after']['info']
        # Verify that the unit and the scaffolding see the same data
        #
        # Scenario presents a secret with a full secret URI to the charm
        # however, the id on the scenario Secret object is a plain id
        assert scenario_secret.id.split('/')[-1] == info['id'].split('/')[-1]
        assert scenario_secret.label == info['label']
        assert scenario_secret._latest_revision == info['revision']
        assert scenario_secret.expire == info['expires']
        assert scenario_secret.rotate == info['rotation']
        assert scenario_secret.description == info['description']
        # https://github.com/canonical/operator/issues/2104
        assert info['rotates'] is None

        assert scenario_secret.tracked_content == result['after']['tracked']
        assert scenario_secret.latest_content == result['after']['latest']



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import os
from unittest.mock import patch

import pytest
from scenario import Context, State
from scenario.errors import UncaughtCharmError
from scenario.state import _Event, _next_action_id

from ops import CharmBase


class MyCharm(CharmBase):
    pass


def test_run():
    ctx = Context(MyCharm, meta={'name': 'foo'})
    state = State()

    with patch.object(ctx, '_run') as p:
        # This would normally be set within the _run call scope.
        ctx._output_state = 'foo'  # type: ignore
        output = ctx.run(ctx.on.start(), state)
        assert output == 'foo'

    assert p.called
    e = p.call_args.kwargs['event']
    s = p.call_args.kwargs['state']

    assert isinstance(e, _Event)
    assert e.name == 'start'
    assert s is state


def test_run_action():
    ctx = Context(MyCharm, meta={'name': 'foo'})
    state = State()
    expected_id = _next_action_id(update=False)

    with patch.object(ctx, '_run') as p:
        # This would normally be set within the _run call scope.
        ctx._output_state = 'foo'  # type: ignore
        output = ctx.run(ctx.on.action('do-foo'), state)
        assert output == 'foo'

    assert p.called
    e = p.call_args.kwargs['event']
    s = p.call_args.kwargs['state']

    assert isinstance(e, _Event)
    assert e.name == 'do_foo_action'
    assert s is state
    assert e.action.id == expected_id


@pytest.mark.parametrize('app_name', ('foo', 'bar', 'george'))
@pytest.mark.parametrize('unit_id', (1, 2, 42))
def test_app_name(app_name, unit_id):
    ctx = Context(MyCharm, meta={'name': 'foo'}, app_name=app_name, unit_id=unit_id)
    with ctx(ctx.on.start(), State()) as mgr:
        assert mgr.charm.app.name == app_name
        assert mgr.charm.unit.name == f'{app_name}/{unit_id}'


@pytest.mark.parametrize('machine_id', ('0', None, '42', '0/lxd/4'))
def test_machine_id_envvar(machine_id):
    ctx = Context(MyCharm, meta={'name': 'foo'}, machine_id=machine_id)
    os.environ.pop('JUJU_MACHINE_ID', None)  # cleanup env to be sure
    with ctx(ctx.on.start(), State()):
        assert os.getenv('JUJU_MACHINE_ID') == machine_id


@pytest.mark.parametrize('availability_zone', ('zone1', None, 'us-east-1a'))
def test_availability_zone_envvar(availability_zone):
    ctx = Context(MyCharm, meta={'name': 'foo'}, availability_zone=availability_zone)
    os.environ.pop('JUJU_AVAILABILITY_ZONE', None)  # cleanup env to be sure
    with ctx(ctx.on.start(), State()):
        assert os.getenv('JUJU_AVAILABILITY_ZONE') == availability_zone


@pytest.mark.parametrize('principal_unit', ('main/0', None, 'app/42'))
def test_principal_unit_envvar(principal_unit):
    ctx = Context(MyCharm, meta={'name': 'foo'}, principal_unit=principal_unit)
    os.environ.pop('JUJU_PRINCIPAL_UNIT', None)  # cleanup env to be sure
    with ctx(ctx.on.start(), State()):
        assert os.getenv('JUJU_PRINCIPAL_UNIT') == principal_unit


def test_context_manager():
    ctx = Context(MyCharm, meta={'name': 'foo'}, actions={'act': {}})
    state = State()
    with ctx(ctx.on.start(), state) as mgr:
        mgr.run()
        assert mgr.charm.meta.name == 'foo'

    with ctx(ctx.on.action('act'), state) as mgr:
        mgr.run()
        assert mgr.charm.meta.name == 'foo'


def test_app_name_and_unit_id_default():
    ctx = Context(MyCharm, meta={'name': 'foo'})
    assert ctx.app_name == 'foo'
    assert ctx.unit_id == 0


def test_app_name_and_unit_id():
    ctx = Context(MyCharm, meta={'name': 'foo'}, app_name='notfoo', unit_id=42)
    assert ctx.app_name == 'notfoo'
    assert ctx.unit_id == 42


@pytest.mark.parametrize('bare_charm_errors', ('1', '0'))
def test_context_manager_uncaught_error(bare_charm_errors: str, monkeypatch: pytest.Monkeypatch):
    class CrashyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.start, self._on_start)
            os.environ['TEST_ENV_VAR'] = '1'

        def _on_start(self, event):
            raise RuntimeError('Crash!')

    monkeypatch.setenv('SCENARIO_BARE_CHARM_ERRORS', bare_charm_errors)
    ctx = Context(CrashyCharm, meta={'name': 'crashy'})
    with pytest.raises((UncaughtCharmError, RuntimeError)):
        with ctx(ctx.on.start(), State()) as mgr:
            assert os.getenv('TEST_ENV_VAR') == '1'
            mgr.run()
    assert 'TEST_ENV_VAR' not in os.environ


@pytest.mark.parametrize('bare_charm_errors', ('1', '0'))
def test_run_uncaught_error(bare_charm_errors: str, monkeypatch: pytest.Monkeypatch):
    class CrashyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.start, self._on_start)
            os.environ['TEST_ENV_VAR'] = '1'

        def _on_start(self, event):
            raise RuntimeError('Crash!')

    monkeypatch.setenv('SCENARIO_BARE_CHARM_ERRORS', bare_charm_errors)
    ctx = Context(CrashyCharm, meta={'name': 'crashy'})
    with pytest.raises((UncaughtCharmError, RuntimeError)):
        ctx.run(ctx.on.start(), State())
    assert 'TEST_ENV_VAR' not in os.environ


def test_context_manager_env_cleared():
    class GoodCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.start, self._on_start)
            os.environ['TEST_ENV_VAR'] = '1'

        def _on_start(self, event):
            os.environ['TEST_ENV_VAR'] = '2'

    ctx = Context(GoodCharm, meta={'name': 'crashy'})
    with ctx(ctx.on.start(), State()) as mgr:
        assert os.getenv('TEST_ENV_VAR') == '1'
        mgr.run()
    assert 'TEST_ENV_VAR' not in os.environ


def test_run_env_cleared():
    class GoodCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.start, self._on_start)

        def _on_start(self, event):
            os.environ['TEST_ENV_VAR'] = '1'

    ctx = Context(GoodCharm, meta={'name': 'crashy'})
    ctx.run(ctx.on.start(), State())
    assert 'TEST_ENV_VAR' not in os.environ


def test_charm_spec_is_deprecated():
    ctx = Context(CharmBase, meta={'name': 'some-name'})
    with pytest.warns(DeprecationWarning):
        _ = ctx.charm_spec  # type: ignore



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import pytest
from scenario import Context, State
from scenario.state import Port, StateValidationError, TCPPort, UDPPort

from ops import CharmBase, Framework, StartEvent, StopEvent


class MyCharm(CharmBase):
    META: Mapping[str, Any] = {'name': 'edgar'}

    def __init__(self, framework: Framework):
        super().__init__(framework)
        framework.observe(self.on.start, self._open_port)
        framework.observe(self.on.stop, self._close_port)

    def _open_port(self, _: StartEvent):
        self.unit.open_port('tcp', 12)

    def _close_port(self, _: StopEvent):
        assert self.unit.opened_ports()
        self.unit.close_port('tcp', 42)


@pytest.fixture
def ctx():
    return Context(MyCharm, meta=MyCharm.META)


def test_open_port(ctx):
    out = ctx.run(ctx.on.start(), State())
    assert len(out.opened_ports) == 1
    port = next(iter(out.opened_ports))

    assert port.protocol == 'tcp'
    assert port.port == 12


def test_close_port(ctx):
    out = ctx.run(ctx.on.stop(), State(opened_ports={TCPPort(42)}))
    assert not out.opened_ports


def test_port_no_arguments():
    with pytest.raises(RuntimeError):
        Port()


@pytest.mark.parametrize('klass', (TCPPort, UDPPort))
def test_port_port(klass):
    with pytest.raises(StateValidationError):
        klass(port=0)
    with pytest.raises(StateValidationError):
        klass(port=65536)



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

from collections.abc import Generator
from typing import Any

import pytest
import yaml
from scenario import Context

from test.charms.test_secrets.src.charm import SecretsCharm


@pytest.fixture
def secrets_context(secrets_charm_meta: dict[str, Any]):
    return Context(SecretsCharm, meta=secrets_charm_meta, actions=secrets_charm_meta['actions'])


@pytest.fixture
def secrets_charm_meta(pytestconfig: pytest.Config) -> Generator[dict[str, Any]]:
    """Metadata for the secret test charm."""
    return yaml.safe_load(
        (pytestconfig.rootpath / 'test/charms/test_secrets/charmcraft.yaml').read_text()
    )



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

from collections.abc import Callable

import pytest
import scenario
from scenario import Context
from scenario.errors import UncaughtCharmError
from scenario.state import (
    _DEFAULT_JUJU_DATABAG,
    PeerRelation,
    Relation,
    RelationBase,
    State,
    StateValidationError,
    SubordinateRelation,
    _Event,
    _next_relation_id,
)

import ops
from ops.charm import (
    CharmBase,
    CharmEvents,
    CollectStatusEvent,
    RelationBrokenEvent,
    RelationCreatedEvent,
    RelationDepartedEvent,
    RelationEvent,
)
from ops.framework import EventBase, Framework
from tests.helpers import trigger


@pytest.fixture(scope='function')
def mycharm():
    class MyCharmEvents(CharmEvents):
        @classmethod
        def define_event(cls, event_kind: str, event_type: type[EventBase]):
            if getattr(cls, event_kind, None):
                delattr(cls, event_kind)
            return super().define_event(event_kind, event_type)

    class MyCharm(CharmBase):
        _call: Callable[[MyCharm, _Event], None] | None = None
        called = False
        on = MyCharmEvents()

        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if self._call:
                MyCharm.called = True
                self._call(event)

    return MyCharm


def test_get_relation(mycharm):
    def pre_event(charm: CharmBase):
        assert charm.model.get_relation('foo')
        assert charm.model.get_relation('bar') is None
        assert charm.model.get_relation('qux')
        assert charm.model.get_relation('zoo') is None

    trigger(
        State(
            config={'foo': 'bar'},
            leader=True,
            relations={
                Relation(endpoint='foo', interface='foo', remote_app_name='remote'),
                Relation(endpoint='qux', interface='qux', remote_app_name='remote'),
            },
        ),
        'start',
        mycharm,
        meta={
            'name': 'local',
            'requires': {
                'foo': {'interface': 'foo'},
                'bar': {'interface': 'bar'},
            },
            'provides': {
                'qux': {'interface': 'qux'},
                'zoo': {'interface': 'zoo'},
            },
        },
        config={'options': {'foo': {'type': 'string'}}},
        pre_event=pre_event,
    )


@pytest.mark.parametrize('test_context', ['init', 'event'])
@pytest.mark.parametrize(
    'is_leader', [pytest.param(True, id='leader'), pytest.param(False, id='minion')]
)
def test_relation_validates_access(is_leader: bool, test_context: str):
    """Test that relation databag read/write access in __init__ is the same as in observers."""

    class Charm(ops.CharmBase):
        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            framework.observe(self.on['my-act'].action, self._on_action)
            self.test_validation('init')

        def _on_action(self, action: ops.ActionEvent):
            self.test_validation('event')

        def test_validation(self, context: str):
            if context != test_context:
                return
            nonlocal validated
            validated += 1
            rel = self.model.get_relation('my-rel')
            assert rel is not None

            # remote application databag
            # any unit can read the remote application databag
            remote_app_data = rel.data[rel.app]
            assert remote_app_data['k'] == 'remote val'
            assert len(remote_app_data.items()) == 1
            # no unit can write to the remote application databag
            with pytest.raises(ops.RelationDataAccessError):
                remote_app_data['k'] = 'something'

            # local application databag
            local_app_data = rel.data[self.app]
            # only the leader can read or write the local application databag
            if self.unit.is_leader():
                assert local_app_data['k'] == 'local val'  # test read
                local_app_data['k'] = 'new val'  # test write
            else:
                with pytest.raises(ops.RelationDataAccessError):
                    local_app_data['k']
            # these probably fail at real runtime with a ModelError
            # but pass here because the validation methods are only hooked up to get/set
            assert len(local_app_data.items()) == 1
            assert 'k' in local_app_data

    ctx = Context(
        Charm,
        meta={
            'name': 'charm',
            'requires': {'my-rel': {'interface': 'my-face'}},
        },
        actions={'my-act': {}},
    )
    rel_in = scenario.Relation(
        endpoint='my-rel',
        local_app_data={'k': 'local val'},
        remote_app_data={'k': 'remote val'},
    )
    validated = 0
    ctx.run(ctx.on.action('my-act'), State(relations={rel_in}, leader=is_leader))
    assert validated


def test_relation_set_single_add_del_change():
    relation_name = 'relation-name'

    class Charm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            framework.observe(self.on.update_status, self._update_status)

        def _update_status(self, event: EventBase):
            rel = self.model.get_relation(relation_name)
            assert rel is not None
            data = rel.data[self.unit]
            data['to-change-key'] = 'to-change-val-new'
            del data['to-remove-key']
            data['new-key'] = 'new-val'

    ctx = Context(
        Charm,
        meta={
            'name': 'charm-name',
            'peers': {relation_name: {'interface': 'interface-name'}},
        },
    )
    rel_in = PeerRelation(
        endpoint=relation_name,
        local_unit_data={
            'to-ignore-key': 'to-ignore-val',
            'to-change-key': 'to-change-val-original',
            'to-remove-key': 'to-remove-val',
        },
    )
    state = ctx.run(ctx.on.update_status(), State(relations={rel_in}))
    rel_out = state.get_relation(rel_in.id)
    assert rel_out.local_unit_data == {
        'to-ignore-key': 'to-ignore-val',
        'to-change-key': 'to-change-val-new',
        'new-key': 'new-val',
    }


@pytest.mark.parametrize(
    ('original_data', 'new_data', 'result_data'),
    [
        pytest.param(
            {},
            {'NEW-KEY-1': 'NEW-VAL-1', 'NEW-KEY-2': 'NEW-VAL-2'},
            {'NEW-KEY-1': 'NEW-VAL-1', 'NEW-KEY-2': 'NEW-VAL-2'},
            id='populate the relation data from scratch',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            {},
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            id='make no changes',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            {'NEW-KEY-1': 'NEW-VAL-1'},
            {
                'old-key-1': 'old-val-1',
                'old-key-2': 'old-val-2',
                'NEW-KEY-1': 'NEW-VAL-1',
            },
            id='insert a new key and value into existing relation data',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1'},
            {'NEW-KEY-1': 'NEW-VAL-1', 'NEW-KEY-2': 'NEW-VAL-2'},
            {
                'old-key-1': 'old-val-1',
                'NEW-KEY-1': 'NEW-VAL-1',
                'NEW-KEY-2': 'NEW-VAL-2',
            },
            id='insert multiple new keys and values',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            {'old-key-1': 'NEW-VAL-1'},
            {'old-key-1': 'NEW-VAL-1', 'old-key-2': 'old-val-2'},
            id='update an existing entry',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            {'old-key-1': 'NEW-VAL-1', 'old-key-2': 'NEW-VAL-2'},
            {'old-key-1': 'NEW-VAL-1', 'old-key-2': 'NEW-VAL-2'},
            id='update multiple existing entries',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            {
                'old-key-1': 'NEW-VAL-1',
                'old-key-2': 'NEW-VAL-2',
                'NEW-KEY-3': 'NEW-VAL-3',
            },
            {
                'old-key-1': 'NEW-VAL-1',
                'old-key-2': 'NEW-VAL-2',
                'NEW-KEY-3': 'NEW-VAL-3',
            },
            id='update multiple existing entries and add a new one',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            {'old-key-1': ''},
            {'old-key-2': 'old-val-2'},
            id='delete an existing entry',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            {'old-key-1': '', 'old-key-2': ''},
            {},
            id='delete multiple existing entries',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            {'NEW-KEY-1': ''},
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            id='deleting a non-existing key has no effect',
        ),
        pytest.param(
            {'old-key-1': 'old-val-1', 'old-key-2': 'old-val-2'},
            {'old-key-1': '', 'old-key-2': '', 'NEW-KEY-1': ''},
            {},
            id='delete multiple existing entries and a non-existing key',
        ),
        pytest.param(
            {
                'old-key-1': 'old-val-1',
                'old-key-2': 'old-val-2',
                'old-key-3': 'old-val-3',
            },
            {'NEW-KEY-1': 'NEW-VAL-1', 'old-key-2': 'NEW-VAL-2', 'old-key-3': ''},
            {
                'old-key-1': 'old-val-1',
                'NEW-KEY-1': 'NEW-VAL-1',
                'old-key-2': 'NEW-VAL-2',
            },
            id='add a key, update another, and delete a third',
        ),
    ],
)
def test_relation_set_bulk_update(
    original_data: dict[str, str], new_data: dict[str, str], result_data: dict[str, str]
):
    relation_name = 'relation-name'

    class Charm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            framework.observe(self.on.update_status, self._update_status)

        def _update_status(self, event: EventBase):
            rel = self.model.get_relation(relation_name)
            assert rel is not None
            data = rel.data[self.unit]
            data.update(new_data)

    ctx = Context(
        Charm,
        meta={
            'name': 'charm-name',
            'peers': {relation_name: {'interface': 'interface-name'}},
        },
    )
    rel_in = PeerRelation(endpoint=relation_name, local_unit_data=original_data)
    state = ctx.run(ctx.on.update_status(), State(relations={rel_in}))
    rel_out = state.get_relation(rel_in.id)
    assert rel_out.local_unit_data == result_data


@pytest.mark.parametrize(
    'evt_name',
    ('changed', 'broken', 'departed', 'joined', 'created'),
)
@pytest.mark.parametrize(
    'remote_app_name',
    ('remote', 'prometheus', 'aodeok123'),
)
def test_relation_events(mycharm, evt_name, remote_app_name):
    relation = Relation(endpoint='foo', interface='foo', remote_app_name=remote_app_name)

    def callback(charm: CharmBase, e):
        if not isinstance(e, RelationEvent):
            return  # filter out collect status events

        if evt_name == 'broken':
            assert charm.model.get_relation('foo') is None
            assert e.relation.app.name == remote_app_name
        else:
            assert charm.model.get_relation('foo').app is not None
            assert charm.model.get_relation('foo').app.name == remote_app_name

    mycharm._call = callback

    trigger(
        State(
            relations={
                relation,
            },
        ),
        f'relation_{evt_name}',
        mycharm,
        meta={
            'name': 'local',
            'requires': {
                'foo': {'interface': 'foo'},
            },
        },
    )


@pytest.mark.parametrize(
    'evt_name,has_unit',
    [
        ('changed', True),
        ('broken', False),
        ('departed', True),
        ('joined', True),
        ('created', False),
    ],
)
@pytest.mark.parametrize(
    'remote_app_name',
    ('remote', 'prometheus', 'aodeok123'),
)
@pytest.mark.parametrize(
    'remote_unit_id',
    (0, 1),
)
def test_relation_events_attrs(mycharm, evt_name, has_unit, remote_app_name, remote_unit_id):
    relation = Relation(endpoint='foo', interface='foo', remote_app_name=remote_app_name)

    def callback(charm: CharmBase, event):
        if isinstance(event, CollectStatusEvent):
            return

        assert event.app
        if not isinstance(event, (RelationCreatedEvent, RelationBrokenEvent)):
            assert event.unit
        if isinstance(event, RelationDepartedEvent):
            assert event.departing_unit

    mycharm._call = callback

    ctx = Context(
        charm_type=mycharm,
        meta={
            'name': 'local',
            'requires': {
                'foo': {'interface': 'foo'},
            },
        },
    )
    state = State(relations={relation})
    kwargs = {}
    if has_unit:
        kwargs['remote_unit'] = remote_unit_id
    event = getattr(ctx.on, f'relation_{evt_name}')(relation, **kwargs)
    ctx.run(event, state=state)


@pytest.mark.parametrize(
    'evt_name',
    ('changed', 'broken', 'departed', 'joined', 'created'),
)
@pytest.mark.parametrize(
    'remote_app_name',
    ('remote', 'prometheus', 'aodeok123'),
)
def test_relation_events_no_attrs(mycharm, evt_name, remote_app_name, caplog):
    relation = Relation(
        endpoint='foo',
        interface='foo',
        remote_app_name=remote_app_name,
        remote_units_data={0: {}, 1: {}},  # 2 units
    )

    def callback(charm: CharmBase, event):
        if isinstance(event, CollectStatusEvent):
            return

        assert event.app  # that's always present
        # .unit is always None for created and broken.
        if isinstance(event, (RelationCreatedEvent, RelationBrokenEvent)):
            assert event.unit is None
        else:
            assert event.unit
        assert (evt_name == 'departed') is bool(getattr(event, 'departing_unit', False))

    mycharm._call = callback

    trigger(
        State(
            relations={
                relation,
            },
        ),
        f'relation_{evt_name}',
        mycharm,
        meta={
            'name': 'local',
            'requires': {
                'foo': {'interface': 'foo'},
            },
        },
    )

    if evt_name not in ('created', 'broken'):
        assert 'remote unit ID unset, and multiple remote unit IDs are present' in caplog.text


def test_relation_default_unit_data_regular():
    relation = Relation('baz')
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.remote_units_data == {0: _DEFAULT_JUJU_DATABAG}


def test_relation_default_unit_data_sub():
    relation = SubordinateRelation('baz')
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.remote_unit_data == _DEFAULT_JUJU_DATABAG


def test_relation_default_unit_data_peer():
    relation = PeerRelation('baz')
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG


@pytest.mark.parametrize('evt_name', ('broken', 'created'))
def test_relation_events_no_remote_units(mycharm, evt_name, caplog):
    relation = Relation(
        endpoint='foo',
        interface='foo',
        remote_units_data={},  # no units
    )

    def callback(charm: CharmBase, event):
        if isinstance(event, CollectStatusEvent):
            return

        assert event.app  # that's always present
        assert not event.unit

    mycharm._call = callback

    trigger(
        State(
            relations={
                relation,
            },
        ),
        f'relation_{evt_name}',
        mycharm,
        meta={
            'name': 'local',
            'requires': {
                'foo': {'interface': 'foo'},
            },
        },
    )

    if evt_name not in ('created', 'broken'):
        assert 'remote unit ID unset; no remote unit data present' in caplog.text


@pytest.mark.parametrize('data', (set(), {}, [], (), 1, 1.0, None, b''))
def test_relation_unit_data_bad_types(mycharm, data):
    with pytest.raises(StateValidationError):
        Relation(endpoint='foo', interface='foo', remote_units_data={0: {'a': data}})


@pytest.mark.parametrize('data', (set(), {}, [], (), 1, 1.0, None, b''))
def test_relation_app_data_bad_types(mycharm, data):
    with pytest.raises(StateValidationError):
        Relation(endpoint='foo', interface='foo', local_app_data={'a': data})


@pytest.mark.parametrize(
    'evt_name',
    ('changed', 'broken', 'departed', 'joined', 'created'),
)
@pytest.mark.parametrize(
    'relation',
    (
        Relation('a', remote_units_data={0: {}}),
        PeerRelation('b', peers_data={1: {}}),
        SubordinateRelation('c'),
    ),
)
def test_relation_event_trigger(relation, evt_name, mycharm):
    meta = {
        'name': 'mycharm',
        'requires': {'a': {'interface': 'i1'}},
        'provides': {
            'c': {
                'interface': 'i3',
                # this is a subordinate relation.
                'scope': 'container',
            }
        },
        'peers': {'b': {'interface': 'i2'}},
    }
    trigger(
        State(relations={relation}),
        f'relation_{evt_name}',
        mycharm,
        meta=meta,
    )


def test_trigger_sub_relation(mycharm):
    meta = {
        'name': 'mycharm',
        'provides': {
            'foo': {
                'interface': 'bar',
                # this is a subordinate relation.
                'scope': 'container',
            }
        },
    }

    sub1 = SubordinateRelation('foo', remote_unit_data={'1': '2'}, remote_app_name='primary1')
    sub2 = SubordinateRelation('foo', remote_unit_data={'3': '4'}, remote_app_name='primary2')

    def post_event(charm: CharmBase):
        b_relations = charm.model.relations['foo']
        assert len(b_relations) == 2
        for relation in b_relations:
            assert len(relation.units) == 1

    trigger(
        State(relations={sub1, sub2}),
        'update_status',
        mycharm,
        meta=meta,
        post_event=post_event,
    )


def test_cannot_instantiate_relationbase():
    with pytest.raises(RuntimeError):
        RelationBase('')


def test_relation_ids():
    from scenario.state import _next_relation_id_counter

    initial_id = _next_relation_id_counter
    for i in range(10):
        rel = Relation('foo')
        assert rel.id == initial_id + i


def test_broken_relation_not_in_model_relations(mycharm):
    rel = Relation('foo')

    ctx = Context(mycharm, meta={'name': 'local', 'requires': {'foo': {'interface': 'foo'}}})
    with ctx(ctx.on.relation_broken(rel), state=State(relations={rel})) as mgr:
        charm = mgr.charm

        assert charm.model.get_relation('foo') is None
        assert charm.model.relations['foo'] == []


def test_get_relation_when_missing():
    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.update_status, self._on_update_status)
            self.framework.observe(self.on.config_changed, self._on_config_changed)
            self.relation = None

        def _on_update_status(self, _):
            self.relation = self.model.get_relation('foo')

        def _on_config_changed(self, _):
            self.relation = self.model.get_relation('foo', self.config['relation-id'])

    ctx = Context(
        MyCharm,
        meta={'name': 'foo', 'requires': {'foo': {'interface': 'foo'}}},
        config={'options': {'relation-id': {'type': 'int', 'description': 'foo'}}},
    )
    # There should be no error if the relation is missing - get_relation returns
    # None in that case.
    with ctx(ctx.on.update_status(), State()) as mgr:
        mgr.run()
        assert mgr.charm.relation is None

    # There should also be no error if the relation is present, of course.
    rel = Relation('foo')
    with ctx(ctx.on.update_status(), State(relations={rel})) as mgr:
        mgr.run()
        assert mgr.charm.relation.id == rel.id

    # If a relation that doesn't exist is requested, that should also not raise
    # an error.
    with ctx(ctx.on.config_changed(), State(config={'relation-id': 42})) as mgr:
        mgr.run()
        rel = mgr.charm.relation
        assert rel.id == 42
        assert not rel.active

    # If there's no defined relation with the name, then get_relation raises KeyError.
    ctx = Context(MyCharm, meta={'name': 'foo'})
    with pytest.raises((KeyError, UncaughtCharmError)) as exc:
        ctx.run(ctx.on.update_status(), State())
    assert isinstance(exc.value, KeyError) or isinstance(exc.value.__cause__, KeyError)


@pytest.mark.parametrize('klass', (Relation, PeerRelation, SubordinateRelation))
def test_relation_positional_arguments(klass):
    with pytest.raises(TypeError):
        klass('foo', 'bar', None)


def test_relation_default_values():
    expected_id = _next_relation_id(update=False)
    endpoint = 'database'
    interface = 'postgresql'
    relation = Relation(endpoint, interface)
    assert relation.id == expected_id
    assert relation.endpoint == endpoint
    assert relation.interface == interface
    assert relation.local_app_data == {}
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.remote_app_name == 'remote'
    assert relation.limit == 1
    assert relation.remote_app_data == {}
    assert relation.remote_units_data == {0: _DEFAULT_JUJU_DATABAG}


def test_subordinate_relation_default_values():
    expected_id = _next_relation_id(update=False)
    endpoint = 'database'
    interface = 'postgresql'
    relation = SubordinateRelation(endpoint, interface)
    assert relation.id == expected_id
    assert relation.endpoint == endpoint
    assert relation.interface == interface
    assert relation.local_app_data == {}
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.remote_app_name == 'remote'
    assert relation.remote_unit_id == 0
    assert relation.remote_app_data == {}
    assert relation.remote_unit_data == _DEFAULT_JUJU_DATABAG


def test_peer_relation_default_values():
    expected_id = _next_relation_id(update=False)
    endpoint = 'peers'
    interface = 'shared'
    relation = PeerRelation(endpoint, interface)
    assert relation.id == expected_id
    assert relation.endpoint == endpoint
    assert relation.interface == interface
    assert relation.local_app_data == {}
    assert relation.local_unit_data == _DEFAULT_JUJU_DATABAG
    assert relation.peers_data == {}


def test_relation_remote_model():
    class MyCharm(CharmBase):
        def __init__(self, framework):
            super().__init__(framework)
            self.framework.observe(self.on.start, self._on_start)

        def _on_start(self, event):
            relation = self.model.get_relation('foo')
            assert relation is not None
            self.remote_model_uuid = relation.remote_model.uuid

    ctx = Context(MyCharm, meta={'name': 'foo', 'requires': {'foo': {'interface': 'foo'}}})
    state = State(relations={Relation('foo')})
    with ctx(ctx.on.start(), state) as mgr:
        mgr.run()
        assert mgr.charm.remote_model_uuid
        assert mgr.charm.remote_model_uuid == mgr.charm.model.uuid

    state = State(relations={Relation('foo', remote_model_uuid='UUID')})
    with ctx(ctx.on.start(), state) as mgr:
        mgr.run()
        assert mgr.charm.remote_model_uuid == 'UUID'
        assert mgr.charm.remote_model_uuid != mgr.charm.model.uuid


def test_peer_relation_units_does_not_contain_this_unit():
    relation_name = 'relation-name'

    class Charm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            framework.observe(self.on.update_status, self._update_status)

        def _update_status(self, _: EventBase):
            rel = self.model.get_relation(relation_name)
            assert rel is not None
            assert self.unit not in rel.units
            data = rel.data[self.unit]
            data['this-unit'] = str(self.unit)

    ctx = Context(
        Charm,
        meta={
            'name': 'charm-name',
            'peers': {relation_name: {'interface': 'interface-name'}},
        },
    )
    rel_in = PeerRelation(
        endpoint=relation_name,
    )
    state = ctx.run(ctx.on.update_status(), State(relations={rel_in}))
    rel_out = state.get_relation(rel_in.id)
    assert rel_out.local_unit_data.get('this-unit') == '<ops.model.Unit charm-name/0>'



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import logging
from collections.abc import Mapping
from typing import Any

import pytest
from scenario import Context
from scenario.state import JujuLogLine, State

from ops.charm import CharmBase, CollectStatusEvent

logger = logging.getLogger('testing logger')


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(CharmBase):
        META: Mapping[str, Any] = {'name': 'mycharm'}

        def __init__(self, framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if isinstance(event, CollectStatusEvent):
                return
            print('foo!')
            logger.warning('bar!')

    return MyCharm


def test_juju_log(mycharm):
    ctx = Context(mycharm, meta=mycharm.META)
    ctx.run(ctx.on.start(), State())
    assert JujuLogLine(level='DEBUG', message='Emitting Juju event start.') in ctx.juju_log
    assert JujuLogLine(level='WARNING', message='bar!') in ctx.juju_log
    # prints are not juju-logged.



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from scenario import Context, State
from scenario.state import _Event

from ops.charm import CharmBase, CharmEvents, CollectStatusEvent, StartEvent
from ops.framework import CommitEvent, EventBase, EventSource, PreCommitEvent


class Foo(EventBase):
    pass


class MyCharmEvents(CharmEvents):
    foo = EventSource(Foo)


class MyCharm(CharmBase):
    META: Mapping[str, Any] = {'name': 'mycharm'}
    on = MyCharmEvents()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.framework.observe(self.on.start, self._on_start)
        self.framework.observe(self.on.foo, self._on_foo)

    def _on_start(self, e):
        self.on.foo.emit()

    def _on_foo(self, e):
        pass


def test_capture_custom_evt_nonspecific_capture_include_fw_evts():
    ctx = Context(MyCharm, meta=MyCharm.META, capture_framework_events=True)
    ctx.run(ctx.on.start(), State())

    emitted = ctx.emitted_events
    assert len(emitted) == 5
    assert isinstance(emitted[0], StartEvent)
    assert isinstance(emitted[1], Foo)
    assert isinstance(emitted[2], CollectStatusEvent)
    assert isinstance(emitted[3], PreCommitEvent)
    assert isinstance(emitted[4], CommitEvent)


def test_capture_juju_evt():
    ctx = Context(MyCharm, meta=MyCharm.META)
    ctx.run(ctx.on.start(), State())

    emitted = ctx.emitted_events
    assert len(emitted) == 2
    assert isinstance(emitted[0], StartEvent)
    assert isinstance(emitted[1], Foo)


def test_capture_deferred_evt():
    ctx = Context(MyCharm, meta=MyCharm.META, capture_deferred_events=True)
    deferred = [_Event('foo').deferred(handler=MyCharm._on_foo)]
    ctx.run(ctx.on.start(), State(deferred=deferred))

    emitted = ctx.emitted_events
    assert len(emitted) == 3
    assert isinstance(emitted[0], Foo)
    assert isinstance(emitted[1], StartEvent)
    assert isinstance(emitted[2], Foo)


def test_capture_no_deferred_evt():
    ctx = Context(MyCharm, meta=MyCharm.META)
    deferred = [_Event('foo').deferred(handler=MyCharm._on_foo)]
    ctx.run(ctx.on.start(), State(deferred=deferred))

    emitted = ctx.emitted_events
    assert len(emitted) == 2
    assert isinstance(emitted[0], StartEvent)
    assert isinstance(emitted[1], Foo)



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import typing
from collections.abc import Mapping

import pytest
from scenario import Context
from scenario.state import Container, Relation, State, _Event

import ops
from ops.framework import LifecycleEvent

from ..helpers import trigger

CHARM_CALLED = 0


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(ops.CharmBase):
        META: Mapping[str, typing.Any] = {
            'name': 'mycharm',
            'requires': {'foo': {'interface': 'bar'}},
            'containers': {'foo': {'type': 'oci-image'}},
        }
        defer_next = 0
        captured: typing.ClassVar[list[ops.EventBase]] = []

        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                if issubclass(evt.event_type, LifecycleEvent):
                    continue

                self.framework.observe(evt, self._on_event)

        def _on_event(self, event: ops.EventBase):
            self.captured.append(event)
            if self.defer_next > 0:
                self.defer_next -= 1
                event.defer()

    return MyCharm


def test_defer(mycharm):
    mycharm.defer_next = True
    out = trigger(State(), 'start', mycharm, meta=mycharm.META)
    assert len(out.deferred) == 1
    assert out.deferred[0].name == 'start'


def test_deferred_evt_emitted(mycharm):
    mycharm.defer_next = 2

    out = trigger(
        State(deferred=[_Event('update_status').deferred(handler=mycharm._on_event)]),
        'start',
        mycharm,
        meta=mycharm.META,
    )

    # we deferred the first 2 events we saw: update-status, start.
    start, update_status = out.deferred
    assert start.name == 'start'
    assert update_status.name == 'update_status'

    # we saw start and update-status.
    upstat, start = mycharm.captured
    assert isinstance(upstat, ops.UpdateStatusEvent)
    assert isinstance(start, ops.StartEvent)


def test_deferred_relation_event(mycharm):
    mycharm.defer_next = 2

    rel = Relation(endpoint='foo', remote_app_name='remote')

    out = trigger(
        State(
            relations={rel},
            deferred=[
                _Event('foo_relation_changed', relation=rel).deferred(
                    handler=mycharm._on_event,
                )
            ],
        ),
        'start',
        mycharm,
        meta=mycharm.META,
    )

    # we deferred the first 2 events we saw: relation-changed, start.
    relation_changed, start = out.deferred
    assert relation_changed.name == 'foo_relation_changed'
    assert start.name == 'start'

    # we saw start and relation-changed.
    relation_changed, start = mycharm.captured
    assert isinstance(relation_changed, ops.RelationChangedEvent)
    assert isinstance(start, ops.StartEvent)


def test_deferred_relation_event_from_relation(mycharm):
    ctx = Context(mycharm, meta=mycharm.META)
    mycharm.defer_next = 2
    rel = Relation(endpoint='foo', remote_app_name='remote')
    out = trigger(
        State(
            relations={rel},
            deferred=[
                ctx.on.relation_changed(rel, remote_unit=1).deferred(handler=mycharm._on_event)
            ],
        ),
        'start',
        mycharm,
        meta=mycharm.META,
    )

    # we deferred the first 2 events we saw: foo_relation_changed, start.
    relation_changed, start = out.deferred
    assert relation_changed.name == 'foo_relation_changed'
    assert relation_changed.snapshot_data == {
        'relation_name': rel.endpoint,
        'relation_id': rel.id,
        'app_name': 'remote',
        'unit_name': 'remote/1',
    }
    assert start.name == 'start'

    # we saw start and foo_relation_changed.
    relation_changed, start = mycharm.captured
    assert isinstance(relation_changed, ops.RelationChangedEvent)
    assert isinstance(start, ops.StartEvent)


def test_deferred_workload_event(mycharm):
    mycharm.defer_next = 2

    ctr = Container('foo')

    out = trigger(
        State(
            containers={ctr},
            deferred=[
                _Event('foo_pebble_ready', container=ctr).deferred(handler=mycharm._on_event)
            ],
        ),
        'start',
        mycharm,
        meta=mycharm.META,
    )

    # we deferred the first 2 events we saw: foo_pebble_ready, start.
    pebble_ready, start = out.deferred
    assert pebble_ready.name == 'foo_pebble_ready'
    assert start.name == 'start'

    # we saw start and foo_pebble_ready.
    ready, start = mycharm.captured
    assert isinstance(ready, ops.WorkloadEvent)
    assert isinstance(start, ops.StartEvent)


def test_defer_reemit_lifecycle_event(mycharm):
    ctx = Context(mycharm, meta=mycharm.META, capture_deferred_events=True)

    mycharm.defer_next = 1
    state_1 = ctx.run(ctx.on.update_status(), State())

    mycharm.defer_next = 0
    state_2 = ctx.run(ctx.on.start(), state_1)

    assert [type(e).__name__ for e in ctx.emitted_events] == [
        'UpdateStatusEvent',
        'UpdateStatusEvent',
        'StartEvent',
    ]
    assert len(state_1.deferred) == 1
    assert not state_2.deferred


def test_defer_reemit_relation_event(mycharm):
    ctx = Context(mycharm, meta=mycharm.META, capture_deferred_events=True)

    rel = Relation('foo')
    mycharm.defer_next = 1
    state_1 = ctx.run(ctx.on.relation_created(rel), State(relations={rel}))

    mycharm.defer_next = 0
    state_2 = ctx.run(ctx.on.start(), state_1)

    assert [type(e).__name__ for e in ctx.emitted_events] == [
        'RelationCreatedEvent',
        'RelationCreatedEvent',
        'StartEvent',
    ]
    assert len(state_1.deferred) == 1
    assert not state_2.deferred


class CustomEventWithArgs(ops.EventBase):
    arg0: str
    arg1: int

    def __init__(self, handle: ops.Handle, arg0: str = '', arg1: int = 0):
        super().__init__(handle)
        self.arg0 = arg0
        self.arg1 = arg1

    def snapshot(self):
        base = super().snapshot()
        base.update({'arg0': self.arg0, 'arg1': self.arg1})
        return base

    def restore(self, snapshot: dict[str, typing.Any]):
        super().restore(snapshot)
        self.arg0 = snapshot['arg0']
        self.arg1 = snapshot['arg1']


class CustomEvents(ops.ObjectEvents):
    foo_changed = ops.EventSource(CustomEventWithArgs)


class MyConsumer(ops.Object):
    on = CustomEvents()  # type: ignore

    def __init__(self, charm: ops.CharmBase):
        super().__init__(charm, 'my-consumer')


def test_defer_custom_event(mycharm):
    class MyCharm(mycharm):
        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            self.consumer = MyConsumer(self)
            framework.observe(self.consumer.on.foo_changed, self._on_event)

    ctx = Context(MyCharm, meta=mycharm.META, capture_deferred_events=True)

    mycharm.defer_next = 1
    state_1 = ctx.run(ctx.on.custom(MyConsumer.on.foo_changed, 'foo', 28), State())
    assert [type(e).__name__ for e in ctx.emitted_events] == ['CustomEventWithArgs']
    assert ctx.emitted_events[0].snapshot() == {'arg0': 'foo', 'arg1': 28}
    assert len(state_1.deferred) == 1

    mycharm.defer_next = 0
    state_2 = ctx.run(ctx.on.start(), state_1)
    assert [type(e).__name__ for e in ctx.emitted_events] == [
        'CustomEventWithArgs',
        'CustomEventWithArgs',
        'StartEvent',
    ]
    assert (
        ctx.emitted_events[0].snapshot()
        == ctx.emitted_events[1].snapshot()
        == {'arg0': 'foo', 'arg1': 28}
    )
    assert not state_2.deferred



# Copyright 2024 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import pathlib

import pytest
from scenario import Context, Resource, State

import ops


class ResourceCharm(ops.CharmBase):
    def __init__(self, framework):
        super().__init__(framework)


def test_get_resource():
    ctx = Context(
        ResourceCharm,
        meta={
            'name': 'resource-charm',
            'resources': {'foo': {'type': 'file'}, 'bar': {'type': 'file'}},
        },
    )
    resource1 = Resource(name='foo', path=pathlib.Path('/var/charm/foo'))
    resource2 = Resource(name='bar', path=pathlib.Path('~/bar'))
    with ctx(ctx.on.update_status(), state=State(resources={resource1, resource2})) as mgr:
        assert mgr.charm.model.resources.fetch('foo') == resource1.path
        assert mgr.charm.model.resources.fetch('bar') == resource2.path
        with pytest.raises(NameError):
            mgr.charm.model.resources.fetch('baz')



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import pytest
from scenario import Context, State, Storage

from ops import CharmBase, ModelError


class MyCharmWithStorage(CharmBase):
    META: Mapping[str, Any] = {
        'name': 'charlene',
        'storage': {'foo': {'type': 'filesystem'}},
    }


class MyCharmWithoutStorage(CharmBase):
    META: Mapping[str, Any] = {'name': 'patrick'}


@pytest.fixture
def storage_ctx():
    return Context(MyCharmWithStorage, meta=MyCharmWithStorage.META)


@pytest.fixture
def no_storage_ctx():
    return Context(MyCharmWithoutStorage, meta=MyCharmWithoutStorage.META)


def test_storage_get_null(no_storage_ctx):
    with no_storage_ctx(no_storage_ctx.on.update_status(), State()) as mgr:
        storages = mgr.charm.model.storages
        assert not len(storages)


def test_storage_get_unknown_name(storage_ctx):
    with storage_ctx(storage_ctx.on.update_status(), State()) as mgr:
        storages = mgr.charm.model.storages
        # not in metadata
        with pytest.raises(KeyError):
            storages['bar']


def test_storage_request_unknown_name(storage_ctx):
    with storage_ctx(storage_ctx.on.update_status(), State()) as mgr:
        storages = mgr.charm.model.storages
        # not in metadata
        with pytest.raises(ModelError):
            storages.request('bar')


def test_storage_get_some(storage_ctx):
    with storage_ctx(storage_ctx.on.update_status(), State()) as mgr:
        storages = mgr.charm.model.storages
        # known but none attached
        assert storages['foo'] == []


@pytest.mark.parametrize('n', (1, 3, 5))
def test_storage_add(storage_ctx, n):
    with storage_ctx(storage_ctx.on.update_status(), State()) as mgr:
        storages = mgr.charm.model.storages
        storages.request('foo', n)

    assert storage_ctx.requested_storages['foo'] == n


def test_storage_usage(storage_ctx):
    storage = Storage('foo')
    # setup storage with some content
    (storage.get_filesystem(storage_ctx) / 'myfile.txt').write_text('helloworld')

    with storage_ctx(storage_ctx.on.update_status(), State(storages={storage})) as mgr:
        foo = mgr.charm.model.storages['foo'][0]
        loc = foo.location
        path = loc / 'myfile.txt'
        assert path.exists()
        assert path.read_text() == 'helloworld'

        myfile = loc / 'path.py'
        myfile.write_text('helloworlds')

    # post-mortem: inspect fs contents.
    assert (storage.get_filesystem(storage_ctx) / 'path.py').read_text() == 'helloworlds'


def test_storage_attached_event(storage_ctx):
    storage = Storage('foo')
    storage_ctx.run(storage_ctx.on.storage_attached(storage), State(storages={storage}))


def test_storage_detaching_event(storage_ctx):
    storage = Storage('foo')
    storage_ctx.run(storage_ctx.on.storage_detaching(storage), State(storages={storage}))



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

from __future__ import annotations

import pytest
from scenario.state import State

from ops.charm import CharmBase
from ops.framework import Framework

from ..helpers import trigger


@pytest.fixture(scope='function')
def mycharm():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            pass

    return MyCharm


def test_config_get(mycharm):
    def check_cfg(charm: CharmBase):
        assert charm.config['foo'] == 'bar'
        assert charm.config['baz'] == 1

    trigger(
        State(
            config={'foo': 'bar', 'baz': 1},
        ),
        'update_status',
        mycharm,
        meta={'name': 'foo'},
        config={'options': {'foo': {'type': 'string'}, 'baz': {'type': 'int'}}},
        post_event=check_cfg,
    )


def test_config_get_default_from_meta(mycharm):
    def check_cfg(charm: CharmBase):
        assert charm.config['foo'] == 'bar'
        assert charm.config['baz'] == 2
        assert charm.config['qux'] is False

    trigger(
        State(
            config={'foo': 'bar'},
        ),
        'update_status',
        mycharm,
        meta={'name': 'foo'},
        config={
            'options': {
                'foo': {'type': 'string'},
                'baz': {'type': 'int', 'default': 2},
                'qux': {'type': 'boolean', 'default': False},
            },
        },
        post_event=check_cfg,
    )


@pytest.mark.parametrize(
    'cfg_in',
    (
        {'foo': 'bar'},
        {'baz': 4, 'foo': 'bar'},
        {'baz': 4, 'foo': 'bar', 'qux': True},
    ),
)
def test_config_in_not_mutated(mycharm, cfg_in):
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            # access the config to trigger a config-get
            foo_cfg = self.config['foo']  # noqa: F841
            baz_cfg = self.config['baz']  # noqa: F841
            qux_cfg = self.config['qux']  # noqa: F841

    state_out = trigger(
        State(
            config=cfg_in,
        ),
        'update_status',
        MyCharm,
        meta={'name': 'foo'},
        config={
            'options': {
                'foo': {'type': 'string'},
                'baz': {'type': 'int', 'default': 2},
                'qux': {'type': 'boolean', 'default': False},
            },
        },
    )
    # check config was not mutated by scenario
    assert state_out.config == cfg_in
